@@ -2,6 +2,7 @@
 
 #import "DebugLogging.h"
 #import "iTermAdvancedSettingsModel.h"
+#import "iTermMalloc.h"
 #import "iTermMetalBufferPool.h"
 #import "iTermMetalDebugInfo.h"
 #import "iTermShaderTypes.h"
@@ -16,7 +17,7 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
 
 + (instancetype)compositeSourceOver {
     iTermMetalBlending *blending = [[iTermMetalBlending alloc] init];
-    // I tried to make this the same as NSCompositeSourceOver. It's not quite right but I have
+    // I tried to make this the same as NSCompositingOperationSourceOver. It's not quite right but I have
     // no idea why.
     blending.rgbBlendOperation = MTLBlendOperationAdd;
     blending.sourceRGBBlendFactor = MTLBlendFactorOne;  // because it's premultiplied
@@ -29,6 +30,24 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
     return blending;
 }
 
+#if ENABLE_TRANSPARENT_METAL_WINDOWS
+
+// See https://en.wikipedia.org/wiki/Alpha_compositing
++ (instancetype)premultipliedCompositing {
+    iTermMetalBlending *blending = [[iTermMetalBlending alloc] init];
+    blending.rgbBlendOperation = MTLBlendOperationAdd;
+    blending.sourceRGBBlendFactor = MTLBlendFactorOne;
+    blending.destinationRGBBlendFactor = MTLBlendFactorOneMinusSourceAlpha;
+    
+    blending.alphaBlendOperation = MTLBlendOperationAdd;
+    blending.sourceAlphaBlendFactor = MTLBlendFactorOne;
+    blending.destinationAlphaBlendFactor = MTLBlendFactorOneMinusSourceAlpha;
+    
+    return blending;
+}
+
+#endif
+
 - (instancetype)init {
     self = [super init];
     if (self) {
@@ -47,11 +66,14 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
 
 @implementation iTermRenderConfiguration
 
-- (instancetype)initWithViewportSize:(vector_uint2)viewportSize scale:(CGFloat)scale {
+- (instancetype)initWithViewportSize:(vector_uint2)viewportSize
+                               scale:(CGFloat)scale
+                  hasBackgroundImage:(BOOL)hasBackgroundImage {
     self = [super init];
     if (self) {
         _viewportSize = viewportSize;
         _scale = scale;
+        _hasBackgroundImage = hasBackgroundImage;
     }
     return self;
 }
@@ -97,7 +119,7 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
     iTermPreciseTimerStatsMeasureAndRecordTimer(stat);
 }
 
-- (iTermPreciseTimerStats *)stats {
+- (nullable iTermPreciseTimerStats *)stats {
     return NULL;
 }
 
@@ -163,9 +185,9 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
 }
 
 - (NSDictionary *)keyForPipelineState {
-    // At the time of writing only the fragment function is mutable. If other inputs to the pipeline
-    // state descriptor become mutable, add them here.
-    return @{ @"fragment function": _fragmentFunctionName ?: @"" };
+    // This must contain all inputs to picking the pipeline state. These can be changed at runtime.
+    return @{ @"fragment function": _fragmentFunctionName ?: @"",
+              @"vertex function": _vertexFunctionName ?: @"" };
 }
 
 - (id<MTLRenderPipelineState>)pipelineState {
@@ -174,7 +196,7 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
         static id<MTLLibrary> defaultLibrary;
         static dispatch_once_t onceToken;
         dispatch_once(&onceToken, ^{
-            defaultLibrary = [_device newDefaultLibrary];
+            defaultLibrary = [self->_device newDefaultLibrary];
         });
         id <MTLFunction> vertexShader = [defaultLibrary newFunctionWithName:_vertexFunctionName];
         ITDebugAssert(vertexShader);
@@ -228,6 +250,7 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
 
 - (void)writeFragmentTexture:(id<MTLTexture>)texture index:(NSUInteger)index toFolder:(NSURL *)folder {
     NSUInteger length = [iTermTexture rawDataSizeForTexture:texture];
+    int samplesPerPixel = [iTermTexture samplesPerPixelForTexture:texture];
     NSMutableData *storage = [NSMutableData dataWithLength:length];
     [texture getBytes:storage.mutableBytes
           bytesPerRow:[iTermTexture bytesPerRowForForTexture:texture]
@@ -236,8 +259,8 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
     NSImage *image = [NSImage imageWithRawData:storage
                                           size:NSMakeSize(texture.width, texture.height)
                                  bitsPerSample:8
-                               samplesPerPixel:4
-                                      hasAlpha:YES
+                               samplesPerPixel:samplesPerPixel
+                                      hasAlpha:samplesPerPixel == 4
                                 colorSpaceName:NSDeviceRGBColorSpace];
     NSString *name = [NSString stringWithFormat:@"texture.%@.%@.png", @(index), texture.label];
     [image saveAsPNGTo:[folder URLByAppendingPathComponent:name].path];
@@ -249,15 +272,15 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
 
 #pragma mark - Protocol Methods
 
-- (__kindof iTermMetalRendererTransientState * _Nonnull)createTransientStateForConfiguration:(iTermRenderConfiguration *)configuration
-                               commandBuffer:(id<MTLCommandBuffer>)commandBuffer {
+- (nullable __kindof iTermMetalRendererTransientState *)createTransientStateForConfiguration:(iTermRenderConfiguration *)configuration
+                                                                               commandBuffer:(id<MTLCommandBuffer>)commandBuffer {
     iTermMetalRendererTransientState *tState = [[self.transientStateClass alloc] initWithConfiguration:configuration];
     tState.pipelineState = [self pipelineState];
     return tState;
 }
 
-- (void)drawWithRenderEncoder:(id<MTLRenderCommandEncoder>)renderEncoder
-               transientState:(NSDictionary *)transientState {
+- (void)drawWithFrameData:(iTermMetalFrameData *)frameData
+           transientState:(NSDictionary *)transientState {
     [self doesNotRecognizeSelector:_cmd];
 }
 
@@ -343,28 +366,50 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
     return pipeline;
 }
 
-- (id<MTLTexture>)textureFromImage:(NSImage *)image context:(nullable iTermMetalBufferPoolContext *)context {
+- (nullable id<MTLTexture>)textureFromImage:(NSImage *)image context:(nullable iTermMetalBufferPoolContext *)context {
     return [self textureFromImage:image context:context pool:nil];
 }
 
-- (id<MTLTexture>)textureFromImage:(NSImage *)image context:(iTermMetalBufferPoolContext *)context pool:(iTermTexturePool *)pool {
+- (void)convertWidth:(NSUInteger)width
+              height:(NSUInteger)height
+             toWidth:(NSUInteger *)widthOut
+              height:(NSUInteger *)heightOut
+        notExceeding:(NSInteger)maxSize {
+    if (height > width) {
+        [self convertWidth:height height:width toWidth:heightOut height:widthOut notExceeding:maxSize];
+        return;
+    }
+    // At this point, width >= height
+    if (width > maxSize) {
+        *widthOut = maxSize;
+        const CGFloat aspectRatio = (CGFloat)height / (CGFloat)width;
+        *heightOut = maxSize * aspectRatio;
+    } else {
+        *widthOut = width;
+        *heightOut = height;
+    }
+}
+
+- (nullable id<MTLTexture>)textureFromImage:(NSImage *)image context:(iTermMetalBufferPoolContext *)context pool:(iTermTexturePool *)pool {
     if (!image) {
         return nil;
     }
-
     NSRect imageRect = NSMakeRect(0, 0, image.size.width, image.size.height);
     CGImageRef imageRef = [image CGImageForProposedRect:&imageRect context:NULL hints:nil];
 
     // Create a suitable bitmap context for extracting the bits of the image
-    NSUInteger width = CGImageGetWidth(imageRef);
-    NSUInteger height = CGImageGetHeight(imageRef);
-
+    NSUInteger width, height;
+    [self convertWidth:CGImageGetWidth(imageRef)
+                height:CGImageGetHeight(imageRef)
+               toWidth:&width
+                height:&height
+          notExceeding:4096];
     if (width == 0 || height == 0) {
         return nil;
     }
 
     CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
-    uint8_t *rawData = (uint8_t *)calloc(height * width * 4, sizeof(uint8_t));
+    uint8_t *rawData = (uint8_t *)iTermCalloc(height * width * 4, sizeof(uint8_t));
     NSUInteger bytesPerPixel = 4;
     NSUInteger bytesPerRow = bytesPerPixel * width;
     NSUInteger bitsPerComponent = 8;
@@ -392,16 +437,19 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
     if (!texture) {
         texture = [_device newTextureWithDescriptor:textureDescriptor];
     }
-    
+
     MTLRegion region = MTLRegionMake2D(0, 0, width, height);
     [texture replaceRegion:region mipmapLevel:0 withBytes:rawData bytesPerRow:bytesPerRow];
 
     [iTermTexture setBytesPerRow:bytesPerRow
                      rawDataSize:height * width * 4
+                 samplesPerPixel:4
                       forTexture:texture];
 
     free(rawData);
-
+    if (texture) {
+        [context didAddTextureOfSize:texture.width * texture.height];
+    }
     return texture;
 }
 
@@ -425,11 +473,12 @@ const NSInteger iTermMetalDriverMaximumNumberOfFramesInFlight = 3;
     iTermMetalDebugDrawInfo *debugDrawInfo = [tState.debugInfo newDrawWithFormatter:self];
     debugDrawInfo.name = NSStringFromClass(tState.class);
 
-    debugDrawInfo.renderPipelineState = tState.pipelineState;
+    debugDrawInfo.fragmentFunctionName = self.fragmentFunctionName;
+    debugDrawInfo.vertexFunctionName = self.vertexFunctionName;
     [renderEncoder setRenderPipelineState:tState.pipelineState];
 
     // Add viewport size to vertex buffers
-    NSDictionary<NSNumber *, id<MTLBuffer>> *vertexBuffers = clientVertexBuffers ?: @{};
+    NSDictionary<NSNumber *, id<MTLBuffer>> *vertexBuffers;
     vertexBuffers =
         [clientVertexBuffers dictionaryBySettingObject:[self vertexBufferForViewportSize:tState.configuration.viewportSize]
                                                 forKey:@(iTermVertexInputIndexViewportSize)];
