@@ -6,11 +6,37 @@
 //
 //
 
+#import "NSAppearance+iTerm.h"
 #import "NSColor+iTerm.h"
 #import "NSImage+iTerm.h"
 
 @implementation NSImage (iTerm)
 
+- (NSImage *)it_imageFillingSize:(NSSize)size {
+    const CGFloat imageAspectRatio = self.size.width / self.size.height;
+    const CGFloat containerAspectRatio = size.width / size.height;
+    NSRect sourceRect;
+    if (imageAspectRatio < containerAspectRatio) {
+        // image is taller than container.
+        sourceRect.origin.x = 0;
+        sourceRect.size.width = self.size.width;
+        sourceRect.size.height = self.size.width / containerAspectRatio;
+        sourceRect.origin.y = (self.size.height - sourceRect.size.height) / 2.0;
+    } else {
+        // container is taller than image
+        sourceRect.origin.y = 0;
+        sourceRect.size.height = self.size.height;
+        sourceRect.size.width = containerAspectRatio * self.size.height;
+        sourceRect.origin.x = (self.size.width - sourceRect.size.width) / 2.0;
+    }
+    return [NSImage imageOfSize:size drawBlock:^{
+        [self drawInRect:NSMakeRect(0, 0, size.width, size.height)
+                fromRect:sourceRect
+               operation:NSCompositingOperationCopy
+                fraction:1];
+    }];
+}
+
 + (NSImage *)imageOfSize:(NSSize)size color:(NSColor *)color {
     return [self imageOfSize:size drawBlock:^{
         [color set];
@@ -18,26 +44,37 @@
     }];
 }
 
-+ (instancetype)imageOfSize:(NSSize)size drawBlock:(void (^)(void))block {
-    NSImage *image = [[[NSImage alloc] initWithSize:size] autorelease];
-
-    [image lockFocus];
-    block();
-    [image unlockFocus];
-
++ (instancetype)imageOfSize:(NSSize)size drawBlock:(void (^ NS_NOESCAPE)(void))block {
+    NSImage *image = [[NSImage alloc] initWithSize:size];
+    [image it_drawWithBlock:block];
     return image;
 }
 
-+ (NSMutableData *)argbDataForImageOfSize:(NSSize)size drawBlock:(void (^)(CGContextRef context))block {
-    NSMutableData *data = [NSMutableData data];
-    CGContextRef context = [NSImage newBitmapContextOfSize:size storage:data];
-    NSGraphicsContext *graphicsContext = [NSGraphicsContext graphicsContextWithCGContext:context flipped:NO];
-    NSGraphicsContext *savedContext = [NSGraphicsContext currentContext];
-    [NSGraphicsContext setCurrentContext:graphicsContext];
-    block(context);
-    [NSGraphicsContext setCurrentContext:savedContext];
-    CGContextRelease(context);
-    return data;
+- (void)it_drawWithBlock:(void (^)(void))block {
+    if (self.size.width == 0 || self.size.height == 0) {
+        return;
+    }
+    [self lockFocus];
+
+    [NSAppearance it_performBlockWithCurrentAppearanceSetToAppearanceForCurrentTheme:^{
+        block();
+    }];
+
+    [self unlockFocus];
+}
+
++ (NSData *)dataWithFourBytesPerPixelFromDataWithOneBytePerPixel:(NSData *)input {
+    NSMutableData *output = [NSMutableData dataWithLength:input.length * 4];
+    unsigned char *ob = (unsigned char *)output.mutableBytes;
+    unsigned char *ib = (unsigned char *)input.bytes;
+    for (int i = 0; i < input.length; i++) {
+        const int j = i * 4;
+        ob[j + 0] = ib[i];
+        ob[j + 1] = ib[i];
+        ob[j + 2] = ib[i];
+        ob[j + 3] = 255;
+    }
+    return output;
 }
 
 + (instancetype)imageWithRawData:(NSData *)data
@@ -46,22 +83,31 @@
                  samplesPerPixel:(NSInteger)samplesPerPixel
                         hasAlpha:(BOOL)hasAlpha
                   colorSpaceName:(NSString *)colorSpaceName {
+    if (samplesPerPixel == 1) {
+        return [self imageWithRawData:[self dataWithFourBytesPerPixelFromDataWithOneBytePerPixel:data]
+                                 size:size
+                        bitsPerSample:8
+                      samplesPerPixel:4
+                             hasAlpha:YES
+                       colorSpaceName:colorSpaceName];
+    }
+    
     assert(data.length == size.width * size.height * bitsPerSample * samplesPerPixel / 8);
     NSBitmapImageRep *bitmapImageRep =
-        [[[NSBitmapImageRep alloc] initWithBitmapDataPlanes:nil  // allocate the pixel buffer for us
-                                                 pixelsWide:size.width
-                                                 pixelsHigh:size.height
-                                              bitsPerSample:bitsPerSample
-                                            samplesPerPixel:samplesPerPixel
-                                                   hasAlpha:hasAlpha
-                                                   isPlanar:NO
-                                             colorSpaceName:colorSpaceName
-                                                bytesPerRow:bitsPerSample * samplesPerPixel * size.width / 8
-                                               bitsPerPixel:bitsPerSample * samplesPerPixel] autorelease];  // 0 means OS infers it
+        [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:nil  // allocate the pixel buffer for us
+                                                pixelsWide:size.width
+                                                pixelsHigh:size.height
+                                             bitsPerSample:bitsPerSample
+                                           samplesPerPixel:samplesPerPixel
+                                                  hasAlpha:hasAlpha
+                                                  isPlanar:NO
+                                            colorSpaceName:colorSpaceName
+                                               bytesPerRow:bitsPerSample * samplesPerPixel * size.width / 8
+                                              bitsPerPixel:bitsPerSample * samplesPerPixel];  // 0 means OS infers it
 
     memmove([bitmapImageRep bitmapData], data.bytes, data.length);
 
-    NSImage *theImage = [[[NSImage alloc] initWithSize:size] autorelease];
+    NSImage *theImage = [[NSImage alloc] initWithSize:size];
     [theImage addRepresentation:bitmapImageRep];
 
     return theImage;
@@ -78,6 +124,10 @@
     return map[type];
 }
 
++ (instancetype)it_imageNamed:(NSImageName)name forClass:(Class)theClass {
+    return [[NSBundle bundleForClass:theClass] imageForResource:name];
+}
+
 - (NSImage *)blurredImageWithRadius:(int)radius {
     // Initially, this used a CIFilter but this doesn't work on some machines for mysterious reasons.
     // Instead, this algorithm implements a really simple box blur. It's quite fast--about 5ms on
@@ -90,7 +140,7 @@
         [image lockFocus];
         [self drawInRect:frame
                 fromRect:frame
-               operation:NSCompositeSourceOver
+               operation:NSCompositingOperationSourceOver
                 fraction:1];
         [image unlockFocus];
         image = [self onePixelBoxBlurOfImage:image alpha:1.0/9.0];
@@ -109,7 +159,7 @@
                                          size.width,
                                          size.height)
                      fromRect:NSMakeRect(0, 0, size.width, size.height)
-                    operation:NSCompositeSourceOver
+                    operation:NSCompositingOperationSourceOver
                      fraction:alpha];
         }
     }
@@ -145,33 +195,33 @@
 }
 
 - (NSImage *)imageWithColor:(NSColor *)color {
-  NSSize size = self.size;
-  NSRect rect = NSZeroRect;
-  rect.size = size;
+    NSSize size = self.size;
+    NSRect rect = NSZeroRect;
+    rect.size = size;
 
-  // Create a bitmap context.
-  NSMutableData *data = [NSMutableData data];
-  CGContextRef context = [self newBitmapContextWithStorage:data];
+    // Create a bitmap context.
+    NSMutableData *data = [NSMutableData data];
+    CGContextRef context = [self newBitmapContextWithStorage:data];
 
-  // Draw myself into that context.
-  CGContextDrawImage(context, rect, [self CGImageForProposedRect:NULL context:nil hints:nil]);
+    // Draw myself into that context.
+    CGContextDrawImage(context, rect, [self CGImageForProposedRect:NULL context:nil hints:nil]);
 
-  // Now draw over it with |color|.
-  CGContextSetFillColorWithColor(context, [color CGColor]);
-  CGContextSetBlendMode(context, kCGBlendModeSourceAtop);
-  CGContextFillRect(context, rect);
+    // Now draw over it with |color|.
+    CGContextSetFillColorWithColor(context, [color CGColor]);
+    CGContextSetBlendMode(context, kCGBlendModeSourceAtop);
+    CGContextFillRect(context, rect);
 
-  // Extract the resulting image into the graphics context.
-  CGImageRef image = CGBitmapContextCreateImage(context);
+    // Extract the resulting image into the graphics context.
+    CGImageRef image = CGBitmapContextCreateImage(context);
 
-  // Convert to NSImage
-  NSImage *coloredImage = [[[NSImage alloc] initWithCGImage:image size:size] autorelease];
+    // Convert to NSImage
+    NSImage *coloredImage = [[NSImage alloc] initWithCGImage:image size:size];
 
-  // Release memory.
-  CGContextRelease(context);
-  CGImageRelease(image);
+    // Release memory.
+    CGContextRelease(context);
+    CGImageRelease(image);
 
-  return coloredImage;
+    return coloredImage;
 }
 
 - (void)saveAsPNGTo:(NSString *)filename {
@@ -183,7 +233,7 @@
     CGImageRef cgImage = [self CGImageForProposedRect:NULL
                                               context:nil
                                                 hints:nil];
-    NSBitmapImageRep *imageRep = [[[NSBitmapImageRep alloc] initWithCGImage:cgImage] autorelease];
+    NSBitmapImageRep *imageRep = [[NSBitmapImageRep alloc] initWithCGImage:cgImage];
     [imageRep setSize:self.size];
     return [imageRep representationUsingType:fileType properties:@{}];
 }
@@ -221,12 +271,12 @@
     [NSGraphicsContext setCurrentContext:ctx];
     [self drawAtPoint:NSZeroPoint
              fromRect:NSZeroRect
-            operation:NSCompositeCopy
+            operation:NSCompositingOperationCopy
              fraction:1.0];
     [ctx flushGraphics];
     [NSGraphicsContext restoreGraphicsState];
-    
-    return [rep autorelease];
+
+    return rep;
 }
 
 - (NSImageRep *)bestRepresentationForScale:(CGFloat)desiredScale {
@@ -246,4 +296,85 @@
     return best;
 }
 
+- (NSImage *)it_imageWithTintColor:(NSColor *)tintColor {
+    if (!tintColor) {
+        return self;
+    }
+    NSSize size = self.size;
+    NSImage *image = [self copy];
+    image.template = NO;
+    [image it_drawWithBlock:^{
+        [tintColor set];
+        NSRectFillUsingOperation(NSMakeRect(0, 0, size.width, size.height),
+                                 NSCompositingOperationSourceAtop);
+    }];
+    return image;
+
+}
+
+- (NSImage *)it_flippedImage {
+    const NSSize size = self.size;
+    if (size.width == 0 || size.height == 0) {
+        return self;
+    }
+    NSAffineTransform *transform = [NSAffineTransform transform];
+    [transform scaleXBy:1 yBy:-1];
+    NSAffineTransform *center = [NSAffineTransform transform];
+    [center translateXBy:size.width / 2. yBy:size.height / 2.];
+    [transform appendTransform:center];
+    NSImage *image = [[NSImage alloc] initWithSize:size];
+    [image lockFocus];
+    [transform concat];
+    NSRect rect = NSMakeRect(0, 0, size.width, size.height);
+    NSPoint corner = NSMakePoint(-size.width / 2., -size.height / 2.);
+    [self drawAtPoint:corner fromRect:rect operation:NSCompositingOperationCopy fraction:1.0];
+    [image unlockFocus];
+    return image;
+}
+
+- (NSImage *)it_imageOfSize:(NSSize)newSize {
+    if (!self.isValid) {
+        return nil;
+    }
+
+    return [NSImage imageOfSize:newSize drawBlock:^{
+        [[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationHigh];
+        [self drawInRect:NSMakeRect(0, 0, newSize.width, newSize.height)
+                fromRect:NSMakeRect(0, 0, self.size.width, self.size.height)
+               operation:NSCompositingOperationCopy
+                fraction:1];
+    }];
+}
+
+- (CGImageRef)CGImage {
+    return [self CGImageForProposedRect:nil context:nil hints:nil];
+}
+
+- (NSImage *)grayscaleImage {
+    const NSSize size = self.size;
+    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceGray();
+    CGContextRef context = CGBitmapContextCreate(nil,
+                                                 size.width,
+                                                 size.height,
+                                                 8,
+                                                 0,
+                                                 colorSpace,
+                                                 kCGImageAlphaNone);
+    const CGRect rect = CGRectMake(0,
+                                   0,
+                                   size.width,
+                                   size.height);
+    CGContextDrawImage(context, rect, self.CGImage);
+    CGImageRef cgImage = CGBitmapContextCreateImage(context);
+
+    NSImage *result = [[NSImage alloc] initWithCGImage:cgImage
+                                                  size:size];
+
+    CFRelease(cgImage);
+    CGContextRelease(context);
+    CGColorSpaceRelease(colorSpace);
+
+    return result;
+}
+
 @end
