@@ -1,9 +1,13 @@
 #import <Cocoa/Cocoa.h>
-#import "FindViewController.h"
 #import "ITAddressBookMgr.h"
 #import "iTerm.h"
+#import "iTermBadgeLabel.h"
 #import "iTermColorMap.h"
+#import "iTermFindDriver.h"
+#import "iTermFocusFollowsMouseController.h"
 #import "iTermIndicatorsHelper.h"
+#import "iTermKeyboardHandler.h"
+#import "iTermLogicalMovementHelper.h"
 #import "iTermSemanticHistoryController.h"
 #import "iTermTextDrawingHelper.h"
 #import "LineBuffer.h"
@@ -16,11 +20,14 @@
 #include <sys/time.h>
 
 @class CRunStorage;
+@class iTermExpect;
 @class iTermFindCursorView;
 @class iTermFindOnPageHelper;
 @class iTermQuickLookController;
 @class iTermSelection;
 @protocol iTermSemanticHistoryControllerDelegate;
+@class iTermURLActionHelper;
+@class iTermVariableScope;
 @class MovingAverage;
 @class PTYScroller;
 @class PTYScrollView;
@@ -33,8 +40,6 @@
 @class VT100Screen;
 @class VT100Terminal;
 
-#define NSLeftAlternateKeyMask  (0x000020 | NSAlternateKeyMask)
-#define NSRightAlternateKeyMask (0x000040 | NSAlternateKeyMask)
 
 // Types of characters. Used when classifying characters for word selection.
 typedef NS_ENUM(NSInteger, PTYCharType) {
@@ -44,39 +49,16 @@ typedef NS_ENUM(NSInteger, PTYCharType) {
     CHARTYPE_OTHER,       // Symbols, etc. Anything that doesn't fall into the other categories.
 };
 
-typedef NS_ENUM(NSInteger, PTYTextViewSelectionEndpoint) {
-    kPTYTextViewSelectionEndpointStart,
-    kPTYTextViewSelectionEndpointEnd
-};
+@protocol PTYTextViewDelegate <NSObject, iTermBadgeLabelDelegate>
 
-typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionDirection) {
-    kPTYTextViewSelectionExtensionDirectionLeft,
-    kPTYTextViewSelectionExtensionDirectionRight,
-
-    // These ignore the unit and are simple movements.
-    kPTYTextViewSelectionExtensionDirectionUp,
-    kPTYTextViewSelectionExtensionDirectionDown,
-    kPTYTextViewSelectionExtensionDirectionStartOfLine,
-    kPTYTextViewSelectionExtensionDirectionEndOfLine,
-    kPTYTextViewSelectionExtensionDirectionTop,
-    kPTYTextViewSelectionExtensionDirectionBottom,
-    kPTYTextViewSelectionExtensionDirectionStartOfIndentation,
-};
-
-typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
-    kPTYTextViewSelectionExtensionUnitCharacter,
-    kPTYTextViewSelectionExtensionUnitWord,
-    kPTYTextViewSelectionExtensionUnitLine,
-    kPTYTextViewSelectionExtensionUnitMark,
-};
-
-@protocol PTYTextViewDelegate <NSObject>
+@property (nonatomic, readonly) NSEdgeInsets textViewEdgeInsets;
 
 - (BOOL)xtermMouseReporting;
 - (BOOL)xtermMouseReportingAllowMouseWheel;
 - (BOOL)isPasting;
 - (void)queueKeyDown:(NSEvent *)event;
 - (void)keyDown:(NSEvent *)event;
+- (void)keyUp:(NSEvent *)event;
 - (BOOL)hasActionableKeyMappingForEvent:(NSEvent *)event;
 - (int)optionKey;
 - (int)rightOptionKey;
@@ -93,6 +75,7 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (void)sendEscapeSequence:(NSString *)text;
 - (void)sendHexCode:(NSString *)codes;
 - (void)sendText:(NSString *)text;
+- (void)sendTextSlowly:(NSString *)text;
 - (void)launchCoprocessWithCommand:(NSString *)command;
 - (void)insertText:(NSString *)string;
 - (PTYTask *)shell;
@@ -108,6 +91,7 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (void)writeTask:(NSString *)string;
 - (void)writeStringWithLatin1Encoding:(NSString *)string;
 - (void)textViewDidBecomeFirstResponder;
+- (void)textViewDidResignFirstResponder;
 - (void)refresh;
 - (BOOL)textViewIsActiveSession;
 - (BOOL)textViewSessionIsBroadcastingInput;
@@ -124,6 +108,8 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (void)textViewCreateTabWithProfileGuid:(NSString *)guid;
 - (void)textViewSelectNextPane;
 - (void)textViewSelectPreviousPane;
+- (void)textViewPasteSpecialWithStringConfiguration:(NSString *)configuration
+                                      fromSelection:(BOOL)fromSelection;
 - (void)textViewEditSession;
 - (void)textViewToggleBroadcastingInput;
 - (void)textViewCloseWithConfirmation;
@@ -133,6 +119,7 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (BOOL)textViewAmbiguousWidthCharsAreDoubleWidth;
 - (PTYScroller *)textViewVerticalScroller;
 - (BOOL)textViewHasCoprocess;
+- (void)textViewStopCoprocess;
 - (void)textViewPostTabContentsChangedNotification;
 - (void)textViewInvalidateRestorableState;
 - (void)textViewDidFindDirtyRects;
@@ -140,7 +127,8 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (void)textViewMovePane;
 - (void)textViewSwapPane;
 - (NSStringEncoding)textViewEncoding;
-- (NSString *)textViewCurrentWorkingDirectory;
+- (void)textViewGetCurrentWorkingDirectoryWithCompletion:(void (^)(NSString *workingDirectory))completion;
+
 - (BOOL)textViewShouldPlaceCursorAt:(VT100GridCoord)coord verticalOk:(BOOL *)verticalOk;
 // If the textview isn't in the key window, the delegate can return YES in this
 // method to cause the cursor to be drawn as though it were key.
@@ -151,7 +139,9 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
                        modifiers:(NSUInteger)modifiers
                           button:(MouseButtonNumber)button
                       coordinate:(VT100GridCoord)coord
-                          deltaY:(CGFloat)deltaY;
+                          deltaY:(CGFloat)deltaY
+        allowDragBeforeMouseDown:(BOOL)allowDragBeforeMouseDown
+                        testOnly:(BOOL)testOnly;
 
 - (VT100GridAbsCoordRange)textViewRangeOfLastCommandOutput;
 - (VT100GridAbsCoordRange)textViewRangeOfCurrentCommand;
@@ -169,6 +159,7 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (BOOL)isRestartable;
 - (void)textViewToggleAnnotations;
 - (BOOL)textViewShouldAcceptKeyDownEvent:(NSEvent *)event;
+- (iTermExpect *)textViewExpect;
 
 // We guess the user is trying to send arrow keys with the scroll wheel in alt screen.
 - (void)textViewThinksUserIsTryingToSendArrowKeysWithScrollWheel:(BOOL)trying;
@@ -185,7 +176,7 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 // Describes the current user, host, and path.
 - (NSURL *)textViewCurrentLocation;
 - (void)textViewBurySession;
-- (void)textViewShowHoverURL:(NSString *)url;
+- (BOOL)textViewShowHoverURL:(NSString *)url;
 
 - (BOOL)textViewCopyMode;
 - (BOOL)textViewCopyModeSelecting;
@@ -195,9 +186,22 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 - (void)textViewNeedsDisplayInRect:(NSRect)rect;
 - (void)textViewDidSelectPasswordPrompt;
 - (NSImage *)textViewBackgroundImage;
-- (BOOL)backgroundImageTiled;
+- (iTermBackgroundImageMode)backgroundImageMode;
 - (BOOL)textViewShouldDrawRect;
-- (void)textViewDidHighightMark;
+- (void)textViewDidHighlightMark;
+- (BOOL)textViewInInteractiveApplication;
+- (BOOL)textViewTerminalStateForMenuItem:(NSMenuItem *)menuItem;
+- (void)textViewToggleTerminalStateForMenuItem:(NSMenuItem *)menuItem;
+- (void)textViewResetTerminal;
+- (CGRect)textViewRelativeFrame;
+- (CGRect)textViewContainerRect;
+- (CGFloat)textViewBadgeTopMargin;
+- (CGFloat)textViewBadgeRightMargin;
+- (iTermVariableScope *)textViewVariablesScope;
+- (BOOL)textViewTerminalBackgroundColorDeterminesWindowDecorationColor;
+- (void)textViewDidUpdateDropTargetVisibility;
+- (void)textViewDidDetectMouseReportingFrustration;
+- (BOOL)textViewCanBury;
 
 @end
 
@@ -209,6 +213,7 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 
 @interface PTYTextView : NSView <
   iTermColorMapDelegate,
+  iTermFocusFollowsMouseFocusReceiver,
   iTermIndicatorsHelperDelegate,
   iTermSemanticHistoryControllerDelegate,
   iTermTextDrawingHelperDelegate,
@@ -248,8 +253,11 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 @property(nonatomic, assign) BOOL asciiLigatures;
 @property(nonatomic, assign) BOOL nonAsciiLigatures;
 
-// Use a bright version of the text color for bold text?
-@property(nonatomic, assign) BOOL useBrightBold;
+// Use the custom bold color
+@property(nonatomic, readonly) BOOL useCustomBoldColor;
+
+// Brighten bold text?
+@property(nonatomic, assign) BOOL brightenBold;
 
 // Ok to render italic text as italics?
 @property(nonatomic, assign) BOOL useItalicFont;
@@ -329,6 +337,10 @@ typedef NS_ENUM(NSInteger, PTYTextViewSelectionExtensionUnit) {
 typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
 @property(nonatomic, copy) PTYTextViewDrawingHookBlock drawingHook;
 
+@property(nonatomic, assign) BOOL showTimestamps;
+
+@property(nonatomic, readonly) BOOL anyAnnotationsAreVisible;
+
 // For tests.
 @property(nonatomic, readonly) NSRect cursorFrame;
 
@@ -360,6 +372,14 @@ typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
 @property (nonatomic, readonly) NSArray<iTermHighlightedRow *> *highlightedRows;
 
 @property (nonatomic) BOOL suppressDrawing;
+@property (nonatomic, readonly) long long firstVisibleAbsoluteLineNumber;
+@property (nonatomic) BOOL useNativePowerlineGlyphs;
+
+@property (nonatomic, readonly) iTermKeyboardHandler *keyboardHandler;
+
+@property (nonatomic, readonly) iTermURLActionHelper *urlActionHelper;
+
+@property (nonatomic, readonly) VT100GridCoord cursorCoord;
 
 // Returns the size of a cell for a given font. hspace and vspace are multipliers and the width
 // and height.
@@ -390,7 +410,7 @@ typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
                              y:(int)y
                             to:(VT100GridWindowedRange *)range
               ignoringNewlines:(BOOL)ignoringNewlines
-                actionRequired:(BOOL)actionRequred
+                actionRequired:(BOOL)actionRequired
                respectDividers:(BOOL)respectDividers;
 
 // Returns range modified by removing nulls (and possibly spaces) from its ends.
@@ -433,13 +453,14 @@ typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
 // Update the scroller color for light or dark backgrounds.
 - (void)updateScrollerForBackgroundColor;
 
-// Remove underline indicating clickable URL.
-- (void)removeUnderline;
+// Remove underline indicating clickable URL. Returns if it changed.
+- (BOOL)removeUnderline;
 
 // Update the scroll position and schedule a redraw. Returns true if anything
 // onscreen is blinking.
 - (BOOL)refresh;
 - (void)setNeedsDisplayOnLine:(int)line;
+- (void)setCursorNeedsDisplay;
 
 // selection
 - (IBAction)selectAll:(id)sender;
@@ -466,7 +487,8 @@ typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
 - (void)findString:(NSString*)aString
   forwardDirection:(BOOL)direction
       mode:(iTermFindMode)mode
-        withOffset:(int)offset;
+        withOffset:(int)offset
+scrollToFirstResult:(BOOL)scrollToFirstResult;
 
 // Remove highlighted terms from previous search.
 // If resetContext is set then the search state will get reset to empty.
@@ -516,13 +538,15 @@ typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
 // Show a visual highlight of a mark on the given line number.
 - (void)highlightMarkOnLine:(int)line hasErrorCode:(BOOL)hasErrorCode;
 
-- (IBAction)installShellIntegration:(id)sender;
-
 // Open a semantic history path.
-- (BOOL)openSemanticHistoryPath:(NSString *)path
+- (void)openSemanticHistoryPath:(NSString *)path
+                  orRawFilename:(NSString *)rawFileName
                workingDirectory:(NSString *)workingDirectory
+                     lineNumber:(NSString *)lineNumber
+                   columnNumber:(NSString *)columnNumber
                          prefix:(NSString *)prefix
-                         suffix:(NSString *)suffix;
+                         suffix:(NSString *)suffix
+                     completion:(void (^)(BOOL ok))completion;
 
 - (PTYFontInfo*)getFontForChar:(UniChar)ch
                      isComplex:(BOOL)isComplex
@@ -568,26 +592,26 @@ typedef void (^PTYTextViewDrawingHookBlock)(iTermTextDrawingHelper *);
                            by:(PTYTextViewSelectionExtensionUnit)unit
                   cursorCoord:(VT100GridCoord)cursorCoord;
 
-- (VT100GridWindowedRange)rangeByExtendingRange:(VT100GridWindowedRange)existingRange
-                                       endpoint:(PTYTextViewSelectionEndpoint)endpoint
-                                      direction:(PTYTextViewSelectionExtensionDirection)direction
-                                      extractor:(iTermTextExtractor *)extractor
-                                           unit:(PTYTextViewSelectionExtensionUnit)unit;
-
-// For focus follows mouse. Allows a new split pane to become focused even though the mouse pointer
-// is elsewhere. Records the mouse position. Refuses first responder as long as the mouse doesn't
-// move.
-- (void)refuseFirstResponderAtCurrentMouseLocation;
+- (iTermLogicalMovementHelper *)logicalMovementHelperForCursorCoordinate:(VT100GridCoord)cursorCoord;
 
 // Undoes -refuseFirstResponderAtCurrentMouseLocation.
 - (void)resetMouseLocationToRefuseFirstResponderAt;
 
 - (void)setTransparencyAffectsOnlyDefaultBackgroundColor:(BOOL)value;
 
+- (IBAction)selectCurrentCommand:(id)sender;
+- (IBAction)selectOutputOfLastCommand:(id)sender;
+
 - (void)showFireworks;
 
 - (BOOL)imageIsVisible:(iTermImageInfo *)image;
 
+// Turns on the flicker fixer (if enabled) while drawing.
+- (void)performBlockWithFlickerFixerGrid:(void (NS_NOESCAPE ^)(void))block;
+
+- (id)contentWithAttributes:(BOOL)attributes;
+- (void)setUseBoldColor:(BOOL)flag brighten:(BOOL)brighten;
+
 #pragma mark - Testing only
 
 - (id)selectedTextAttributed:(BOOL)attributed
