@@ -7,15 +7,20 @@
 //
 
 #import "ToolProfiles.h"
-#import "PseudoTerminal.h"
+
+#import "DebugLogging.h"
 #import "iTermController.h"
+#import "iTermSessionLauncher.h"
+#import "NSEvent+iTerm.h"
 #import "ProfileModel.h"
+#import "PseudoTerminal.h"
 
 static const int kVerticalMargin = 5;
 static const int kMargin = 0;
 static const int kPopupHeight = 26;
 static const CGFloat kButtonHeight = 23;
 static const CGFloat kInnerMargin = 5;
+static NSString *const iTermToolProfilesProfileListViewState = @"iTermToolProfilesProfileListViewState";
 
 @implementation ToolProfiles {
     ProfileListView *listView_;
@@ -34,11 +39,13 @@ static const CGFloat kInnerMargin = 5;
         [listView_ disableArrowHandler];
         [listView_ allowMultipleSelections];
         [listView_.tableView setHeaderView:nil];
+        if (@available(macOS 10.14, *)) { } else {
+            listView_.tableView.enclosingScrollView.drawsBackground = NO;
+        }
 
         [self addSubview:listView_];
-        [listView_ release];
 
-        _openButton = [[[NSButton alloc] initWithFrame:NSMakeRect(0, frame.size.height - kButtonHeight, frame.size.width, kButtonHeight)] autorelease];
+        _openButton = [[NSButton alloc] initWithFrame:NSMakeRect(0, frame.size.height - kButtonHeight, frame.size.width, kButtonHeight)];
         [_openButton setButtonType:NSMomentaryPushInButton];
         [_openButton setTitle:@"Open"];
         [_openButton setTarget:self];
@@ -49,8 +56,8 @@ static const CGFloat kInnerMargin = 5;
         [self addSubview:_openButton];
         [_openButton bind:@"enabled" toObject:listView_ withKeyPath:@"hasSelection" options:nil];
 
-        popup_ = [[[NSPopUpButton alloc] initWithFrame:NSMakeRect(0, frame.size.height - kPopupHeight, frame.size.width - _openButton.frame.size.width - kInnerMargin, kPopupHeight)] autorelease];
-        [[popup_ cell] setControlSize:NSSmallControlSize];
+        popup_ = [[NSPopUpButton alloc] initWithFrame:NSMakeRect(0, frame.size.height - kPopupHeight, frame.size.width - _openButton.frame.size.width - kInnerMargin, kPopupHeight)];
+        [[popup_ cell] setControlSize:NSControlSizeSmall];
         [[popup_ cell] setFont:[NSFont systemFontOfSize:[NSFont smallSystemFontSize]]];
         [[popup_ menu] addItemWithTitle:@"New Tab"
                                  action:@selector(toolProfilesNewTab:)
@@ -69,6 +76,10 @@ static const CGFloat kInnerMargin = 5;
         }
         [self addSubview:popup_];
         [popup_ setAutoresizingMask:NSViewMinYMargin | NSViewWidthSizable];
+        [[NSNotificationCenter defaultCenter] addObserver:self
+                                                 selector:@selector(refreshTerminal:)
+                                                     name:kRefreshTerminalNotification
+                                                   object:nil];
 
         [popup_ bind:@"enabled" toObject:listView_ withKeyPath:@"hasSelection" options:nil];
 
@@ -76,11 +87,18 @@ static const CGFloat kInnerMargin = 5;
     return self;
 }
 
-- (void)dealloc
-{
+- (void)dealloc {
+    [[NSNotificationCenter defaultCenter] removeObserver:self];
     [popup_ unbind:@"enabled"];
     [_openButton unbind:@"enabled"];
-    [super dealloc];
+}
+
+- (void)refreshTerminal:(NSNotification *)notification {
+    [listView_ reloadData];
+}
+
+- (void)windowBackgroundColorDidChange {
+    [listView_ reloadData];
 }
 
 - (void)relayout {
@@ -103,44 +121,66 @@ static const CGFloat kInnerMargin = 5;
     PseudoTerminal* terminal = [[iTermController sharedInstance] currentTerminal];
     for (NSString* guid in [listView_ selectedGuids]) {
         Profile* bookmark = [[ProfileModel sharedInstance] bookmarkWithGuid:guid];
-        [[iTermController sharedInstance] launchBookmark:bookmark
-                                              inTerminal:terminal];
-    }    
+        [iTermSessionLauncher launchBookmark:bookmark
+                                  inTerminal:terminal
+                          respectTabbingMode:NO
+                                  completion:nil];
+    }
 }
 
 - (void)toolProfilesNewWindow:(id)sender
 {
     for (NSString* guid in [listView_ selectedGuids]) {
         Profile* bookmark = [[ProfileModel sharedInstance] bookmarkWithGuid:guid];
-        [[iTermController sharedInstance] launchBookmark:bookmark
-                                              inTerminal:nil];
-    }    
+        [iTermSessionLauncher launchBookmark:bookmark
+                                  inTerminal:nil
+                          respectTabbingMode:NO
+                                  completion:nil];
+    }
 }
 
 - (void)toolProfilesNewHorizontalSplit:(id)sender
 {
     PseudoTerminal* terminal = [[iTermController sharedInstance] currentTerminal];
     for (NSString* guid in [listView_ selectedGuids]) {
-        [terminal splitVertically:NO withBookmarkGuid:guid];
-    }    
+        Profile *profile = [[ProfileModel sharedInstance] bookmarkWithGuid:guid];
+        if (!profile) {
+            continue;
+        }
+        [terminal asyncSplitVertically:NO
+                                before:NO
+                               profile:profile
+                         targetSession:[terminal currentSession]
+                            completion:nil
+                                 ready:nil];
+    }
 }
 
 - (void)toolProfilesNewVerticalSplit:(id)sender
 {
     PseudoTerminal* terminal = [[iTermController sharedInstance] currentTerminal];
     for (NSString* guid in [listView_ selectedGuids]) {
-        [terminal splitVertically:YES withBookmarkGuid:guid];
+        Profile *profile = [[ProfileModel sharedInstance] bookmarkWithGuid:guid];
+        if (!profile) {
+            continue;
+        }
+        [terminal asyncSplitVertically:YES
+                                before:NO
+                               profile:profile
+                         targetSession:[terminal currentSession]
+                            completion:nil
+                                 ready:nil];
     }
 }
 
 - (void)profileTableRowSelected:(id)profileTable
 {
     NSEvent *event = [[NSApplication sharedApplication] currentEvent];
-    if ([event modifierFlags] & (NSControlKeyMask)) {
+    if ([event it_modifierFlags] & (NSEventModifierFlagControl)) {
         [self toolProfilesNewHorizontalSplit:nil];
-    } else if ([event modifierFlags] & (NSAlternateKeyMask)) {
+    } else if ([event it_modifierFlags] & (NSEventModifierFlagOption)) {
         [self toolProfilesNewVerticalSplit:nil];
-    } else if ([event modifierFlags] & (NSShiftKeyMask)) {
+    } else if ([event it_modifierFlags] & (NSEventModifierFlagShift)) {
         [self toolProfilesNewWindow:nil];
     } else {
         [self toolProfilesNewTab:nil];
@@ -157,7 +197,16 @@ static const CGFloat kInnerMargin = 5;
 }
 
 - (void)open:(id)sender {
-    [self performSelector:[[popup_ selectedItem] action]];
+    [self it_performNonObjectReturningSelector:[[popup_ selectedItem] action]
+                                    withObject:nil];
+}
+
+- (NSDictionary *)restorableState {
+    return @{ iTermToolProfilesProfileListViewState: listView_.restorableState };
+}
+
+- (void)restoreFromState:(NSDictionary *)state {
+    [listView_ restoreFromState:state[iTermToolProfilesProfileListViewState]];
 }
 
 @end
