@@ -8,17 +8,21 @@
 #import "TriggerController.h"
 
 #import "AlertTrigger.h"
+#import "AnnotateTrigger.h"
 #import "BellTrigger.h"
 #import "BounceTrigger.h"
 #import "CaptureTrigger.h"
 #import "CoprocessTrigger.h"
 #import "DebugLogging.h"
 #import "FutureMethods.h"
-#import "GrowlTrigger.h"
+#import "iTermUserNotificationTrigger.h"
 #import "HighlightTrigger.h"
 #import "iTermSetTitleTrigger.h"
 #import "ITAddressBookMgr.h"
+#import "iTermFunctionCallTextFieldDelegate.h"
 #import "iTermNoColorAccessoryButton.h"
+#import "iTermProfilePreferences.h"
+#import "iTermRPCTrigger.h"
 #import "iTermShellPromptTrigger.h"
 #import "MarkTrigger.h"
 #import "NSColor+iTerm.h"
@@ -29,6 +33,7 @@
 #import "SetDirectoryTrigger.h"
 #import "SetHostnameTrigger.h"
 #import "StopTrigger.h"
+#import "iTermHyperlinkTrigger.h"
 #import "Trigger.h"
 
 #import <ColorPicker/ColorPicker.h>
@@ -67,13 +72,14 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
 // Keeps the color well whose popover is currently open from getting
 // deallocated. It may get removed from the view hierarchy but we need it to
 // continue existing so we can get the color out of it.
-@property(nonatomic, retain) iTermColorWell *activeWell;
+@property(nonatomic, strong) iTermColorWell *activeWell;
 @end
 
 @implementation TriggerController {
     NSArray *_triggers;
     // Gives the index of the row being edited while a textfield cell is editing.
     NSInteger _textEditingRow;
+    id _parameterDelegate;
 
     IBOutlet NSTableView *_tableView;
     IBOutlet NSTableColumn *_regexColumn;
@@ -81,6 +87,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
     IBOutlet NSTableColumn *_actionColumn;
     IBOutlet NSTableColumn *_parametersColumn;
     IBOutlet NSButton *_removeTriggerButton;
+    IBOutlet NSButton *_interpolatedStringParameters;
 }
 
 - (instancetype)init {
@@ -88,25 +95,22 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
     if (self) {
         NSMutableArray *triggers = [NSMutableArray array];
         for (Class class in [self triggerClasses]) {
-            [triggers addObject:[[[class alloc] init] autorelease]];
+            [triggers addObject:[[class alloc] init]];
         }
-        _triggers = [triggers retain];
+        _triggers = triggers;
+        _textEditingRow = -1;
     }
     return self;
 }
 
-- (void)dealloc {
-    [_guid release];
-    [_triggers release];
-    [super dealloc];
-}
-
 - (NSArray *)triggerClasses {
     NSArray *allClasses = @[ [AlertTrigger class],
+                             [AnnotateTrigger class],
                              [BellTrigger class],
                              [BounceTrigger class],
+                             [iTermRPCTrigger class],
                              [CaptureTrigger class],
-                             [GrowlTrigger class],
+                             [iTermUserNotificationTrigger class],
                              [iTermShellPromptTrigger class],
                              [iTermSetTitleTrigger class],
                              [SendTextTrigger class],
@@ -116,6 +120,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                              [HighlightTrigger class],
                              [MarkTrigger class],
                              [PasswordTrigger class],
+                             [iTermHyperlinkTrigger class],
                              [SetDirectoryTrigger class],
                              [SetHostnameTrigger class],
                              [StopTrigger class] ];
@@ -193,7 +198,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
         // Stop editing. A reload while editing crashes.
         [_tableView reloadData];
     }
-    NSMutableArray *triggerDictionaries = [[[self triggerDictionariesForCurrentProfile] mutableCopy] autorelease];
+    NSMutableArray *triggerDictionaries = [[self triggerDictionariesForCurrentProfile] mutableCopy];
     if (rowIndex < 0) {
         assert(triggerDictionary);
         [triggerDictionaries addObject:triggerDictionary];
@@ -213,11 +218,11 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
 - (void)moveTriggerOnRow:(int)sourceRow toRow:(int)destinationRow {
     // Stop editing. A reload while editing crashes.
     [_tableView reloadData];
-    NSMutableArray *triggerDictionaries = [[[self triggerDictionariesForCurrentProfile] mutableCopy] autorelease];
+    NSMutableArray *triggerDictionaries = [[self triggerDictionariesForCurrentProfile] mutableCopy];
     if (destinationRow > sourceRow) {
         --destinationRow;
     }
-    NSDictionary *temp = [[triggerDictionaries[sourceRow] retain] autorelease];
+    NSDictionary *temp = triggerDictionaries[sourceRow];
     [triggerDictionaries removeObjectAtIndex:sourceRow];
     [triggerDictionaries insertObject:temp atIndex:destinationRow];
     [_delegate triggerChanged:self newValue:triggerDictionaries];
@@ -235,39 +240,23 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
               kTriggerActionKey: [trigger action] };
 }
 
-- (IBAction)addTrigger:(id)sender {
-    [self setTriggerDictionary:[self defaultTriggerDictionary] forRow:-1 reloadData:YES];
-    [_tableView selectRowIndexes:[NSIndexSet indexSetWithIndex:_tableView.numberOfRows - 1]
-            byExtendingSelection:NO];
-}
-
-- (IBAction)removeTrigger:(id)sender {
-    if (_tableView.selectedRow < 0) {
-        XLog(@"This shouldn't happen: you pressed the button to remove a trigger but no row is selected");
-        return;
-    }
-    [self setTriggerDictionary:nil forRow:[_tableView selectedRow] reloadData:YES];
-    self.hasSelection = [_tableView numberOfSelectedRows] > 0;
-    _removeTriggerButton.enabled = self.hasSelection;
-}
-
 - (void)setGuid:(NSString *)guid {
-    [_guid autorelease];
     _guid = [guid copy];
     [_tableView reloadData];
+    _interpolatedStringParameters.state = [iTermProfilePreferences boolForKey:KEY_TRIGGERS_USE_INTERPOLATED_STRINGS inProfile:[self bookmark]] ? NSOnState : NSOffState;
 }
 
 - (NSTextField *)labelWithString:(NSString *)string origin:(NSPoint)origin {
-    NSTextField *textField = [[[NSTextField alloc] initWithFrame:NSMakeRect(origin.x,
-                                                                            origin.y,
-                                                                            0,
-                                                                            0)] autorelease];
+    NSTextField *textField = [[NSTextField alloc] initWithFrame:NSMakeRect(origin.x,
+                                                                           origin.y,
+                                                                           0,
+                                                                           0)];
     [textField setBezeled:NO];
     [textField setDrawsBackground:NO];
     [textField setEditable:NO];
     [textField setSelectable:NO];
     textField.font = [NSFont systemFontOfSize:[NSFont systemFontSize]];
-    textField.textColor = [NSColor blackColor];
+    textField.textColor = [NSColor textColor];
     textField.stringValue = string;
     [textField sizeToFit];
 
@@ -353,7 +342,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                   row:(NSInteger)row {
     NSDictionary *triggerDictionary = [self triggerDictionariesForCurrentProfile][row];
     if (tableColumn == _actionColumn) {
-        NSPopUpButton *popUpButton = [[[NSPopUpButton alloc] init] autorelease];
+        NSPopUpButton *popUpButton = [[NSPopUpButton alloc] init];
         [popUpButton setTitle:[[_triggers[0] class] title]];
         popUpButton.bordered = NO;
         for (int i = 0; i < [self numberOfTriggers]; i++) {
@@ -368,10 +357,10 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
     } else if (tableColumn == _regexColumn) {
         NSDictionary *triggerDictionary = [self triggerDictionariesForCurrentProfile][row];
         NSTextField *textField =
-            [[[NSTextField alloc] initWithFrame:NSMakeRect(0,
-                                                           0,
-                                                           tableColumn.width,
-                                                           self.tableView.rowHeight)] autorelease];
+            [[NSTextField alloc] initWithFrame:NSMakeRect(0,
+                                                          0,
+                                                          tableColumn.width,
+                                                          self.tableView.rowHeight)];
         textField.font = [NSFont systemFontOfSize:[NSFont systemFontSize]];
         textField.stringValue = triggerDictionary[kTriggerRegexKey] ?: @"";
         textField.editable = YES;
@@ -383,7 +372,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
 
         return textField;
     } else if (tableColumn == _partialLineColumn) {
-        NSButton *checkbox = [[[NSButton alloc] initWithFrame:NSZeroRect] autorelease];
+        NSButton *checkbox = [[NSButton alloc] initWithFrame:NSZeroRect];
         [checkbox sizeToFit];
         [checkbox setButtonType:NSSwitchButton];
         checkbox.title = @"";
@@ -397,20 +386,20 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
         trigger.param = triggerDicts[row][kTriggerParameterKey];
         if ([trigger takesParameter]) {
             if ([trigger paramIsTwoColorWells]) {
-                NSView *container = [[[NSView alloc] initWithFrame:NSMakeRect(0,
-                                                                              0,
-                                                                              tableColumn.width,
-                                                                              _tableView.rowHeight)] autorelease];
+                NSView *container = [[NSView alloc] initWithFrame:NSMakeRect(0,
+                                                                             0,
+                                                                             tableColumn.width,
+                                                                             _tableView.rowHeight)];
                 CGFloat x = 4;
                 NSTextField *label = [self labelWithString:@"Text:" origin:NSMakePoint(x, 0)];
                 [container addSubview:label];
                 x += label.frame.size.width;
                 const CGFloat kWellWidth = 30;
                 iTermColorWell *well =
-                    [[[iTermColorWell alloc] initWithFrame:NSMakeRect(x,
-                                                                      0,
-                                                                      kWellWidth,
-                                                                      _tableView.rowHeight)] autorelease];
+                    [[iTermColorWell alloc] initWithFrame:NSMakeRect(x,
+                                                                     0,
+                                                                     kWellWidth,
+                                                                     _tableView.rowHeight)];
                 well.noColorAllowed = YES;
                 well.continuous = NO;
                 well.tag = row;
@@ -420,8 +409,12 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                 well.target = self;
                 well.action = @selector(colorWellDidChange:);
                 well.identifier = kTextColorWellIdentifier;
+                __weak __typeof(self) weakSelf = self;
+                __weak __typeof(well) weakWell = well;
                 well.willOpenPopover = ^() {
-                    self.activeWell = well;
+                    if (weakWell) {
+                        weakSelf.activeWell = weakWell;
+                    }
                 };
                 well.willClosePopover = ^() {
                     if (self.activeWell == well) {
@@ -432,10 +425,10 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                 label = [self labelWithString:@"Background:" origin:NSMakePoint(x, 0)];
                 [container addSubview:label];
                 x += label.frame.size.width;
-                well = [[[iTermColorWell alloc] initWithFrame:NSMakeRect(x,
-                                                                         0,
-                                                                         kWellWidth,
-                                                                         _tableView.rowHeight)] autorelease];
+                well = [[iTermColorWell alloc] initWithFrame:NSMakeRect(x,
+                                                                        0,
+                                                                        kWellWidth,
+                                                                        _tableView.rowHeight)];
                 well.noColorAllowed = YES;
                 well.continuous = NO;
                 well.color = trigger.backgroundColor;
@@ -444,8 +437,11 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                 well.target = self;
                 well.action = @selector(colorWellDidChange:);
                 well.identifier = kBackgroundColorWellIdentifier;
+                weakWell = well;
                 well.willOpenPopover = ^() {
-                    self.activeWell = well;
+                    if (weakWell) {
+                        weakSelf.activeWell = weakWell;
+                    }
                 };
                 well.willClosePopover = ^() {
                     if (self.activeWell == well) {
@@ -455,7 +451,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
 
                 return container;
             } else if ([trigger paramIsPopupButton]) {
-                NSPopUpButton *popUpButton = [[[NSPopUpButton alloc] init] autorelease];
+                NSPopUpButton *popUpButton = [[NSPopUpButton alloc] init];
                 [popUpButton setTitle:@""];
                 popUpButton.bordered = NO;
 
@@ -469,9 +465,9 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                     for (id object in [trigger objectsSortedByValueInDict:items]) {
                         NSString *theTitle = [items objectForKey:object];
                         if (theTitle) {
-                            NSMenuItem *anItem = [[[NSMenuItem alloc] initWithTitle:theTitle
-                                                                             action:nil
-                                                                      keyEquivalent:@""] autorelease];
+                            NSMenuItem *anItem = [[NSMenuItem alloc] initWithTitle:theTitle
+                                                                            action:nil
+                                                                     keyEquivalent:@""];
                             [theMenu addItem:anItem];
                         }
                     }
@@ -490,27 +486,28 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
                 return popUpButton;
             } else {
                 // If not a popup button, then text by default.
-                NSTextField *textField =
-                    [[[NSTextField alloc] initWithFrame:NSMakeRect(0,
-                                                                   0,
-                                                                   tableColumn.width,
-                                                                   self.tableView.rowHeight)] autorelease];
+                iTermFocusReportingTextField *textField =
+                    [[iTermFocusReportingTextField alloc] initWithFrame:NSMakeRect(0,
+                                                                                   0,
+                                                                                   tableColumn.width,
+                                                                                   self.tableView.rowHeight)];
                 textField.font = [NSFont systemFontOfSize:[NSFont systemFontSize]];
                 textField.stringValue = triggerDictionary[kTriggerParameterKey] ?: @"";
                 textField.editable = YES;
                 textField.selectable = YES;
                 textField.bordered = NO;
                 textField.drawsBackground = NO;
-                textField.delegate = self;
+                _parameterDelegate = [trigger newParameterDelegateWithPassthrough:self];
+                textField.delegate = _parameterDelegate ?: self;
                 if ([textField respondsToSelector:@selector(setPlaceholderString:)]) {
-                    textField.placeholderString = [trigger paramPlaceholder];
+                    textField.placeholderString = [trigger triggerOptionalParameterPlaceholderWithInterpolation:_interpolatedStringParameters.state == NSOnState];
                 }
                 textField.identifier = kParameterColumnIdentifier;
 
                 return textField;
             }
         } else {
-            return [[[NSView alloc] initWithFrame:NSZeroRect] autorelease];
+            return [[NSView alloc] initWithFrame:NSZeroRect];
         }
     }
     return nil;
@@ -533,6 +530,28 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
 
 #pragma mark - Actions
 
+- (IBAction)addTrigger:(id)sender {
+    [self setTriggerDictionary:[self defaultTriggerDictionary] forRow:-1 reloadData:YES];
+    [_tableView selectRowIndexes:[NSIndexSet indexSetWithIndex:_tableView.numberOfRows - 1]
+            byExtendingSelection:NO];
+}
+
+- (IBAction)removeTrigger:(id)sender {
+    if (_tableView.selectedRow < 0) {
+        XLog(@"This shouldn't happen: you pressed the button to remove a trigger but no row is selected");
+        return;
+    }
+    [self setTriggerDictionary:nil forRow:[_tableView selectedRow] reloadData:YES];
+    self.hasSelection = [_tableView numberOfSelectedRows] > 0;
+    _removeTriggerButton.enabled = self.hasSelection;
+}
+
+- (IBAction)toggleUseInterpolatedStrings:(id)sender {
+    _interpolatedStringParameters.state = (![iTermProfilePreferences boolForKey:KEY_TRIGGERS_USE_INTERPOLATED_STRINGS inProfile:[self bookmark]]) ? NSOnState : NSOffState;
+    [self.delegate triggerSetUseInterpolatedStrings:_interpolatedStringParameters.state == NSOnState];
+    [_tableView reloadData];
+}
+
 - (void)doubleClick:(id)sender {
     NSPoint screenLocation = [NSEvent mouseLocation];
     NSPoint windowLocation = [self.window convertRectFromScreen:NSMakeRect(screenLocation.x,
@@ -561,7 +580,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
         return;
     }
     NSMutableDictionary *triggerDictionary =
-        [[[self triggerDictionariesForCurrentProfile][row] mutableCopy] autorelease];
+        [[self triggerDictionariesForCurrentProfile][row] mutableCopy];
     HighlightTrigger *trigger =
         (HighlightTrigger *)[HighlightTrigger triggerFromDict:triggerDictionary];
     if ([colorWell.identifier isEqual:kTextColorWellIdentifier]) {
@@ -583,16 +602,16 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
 - (void)instantDidChange:(NSButton *)checkbox {
     NSNumber *newValue = checkbox.state == NSOnState ? @(YES) : @(NO);
     NSInteger row = [_tableView rowForView:checkbox];
-    
+
     // If a text field is editing, make it save its contents before we get the trigger dictionary.
     [_tableView reloadData];
-    
+
     NSArray *triggerDicts = [self triggerDictionariesForCurrentProfile];
     if (row < 0 || row >= triggerDicts.count) {
         return;
     }
     NSMutableDictionary *triggerDictionary =
-        [[[self triggerDictionariesForCurrentProfile][row] mutableCopy] autorelease];
+        [[self triggerDictionariesForCurrentProfile][row] mutableCopy];
     triggerDictionary[kTriggerPartialLineKey] = newValue;
     [self setTriggerDictionary:triggerDictionary forRow:row reloadData:YES];
 }
@@ -603,18 +622,20 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
         return;
     }
     NSInteger indexOfSelectedAction = [sender indexOfSelectedItem];
-    
+
     // If a text field is being edited, end it and update the trigger dictionary before we fetch it.
     [_tableView reloadData];
 
     NSMutableDictionary *triggerDictionary =
-        [[[self triggerDictionariesForCurrentProfile][rowIndex] mutableCopy] autorelease];
+        [[self triggerDictionariesForCurrentProfile][rowIndex] mutableCopy];
     Trigger *theTrigger = _triggers[indexOfSelectedAction];
     triggerDictionary[kTriggerActionKey] = [theTrigger action];
     [triggerDictionary removeObjectForKey:kTriggerParameterKey];
     Trigger *triggerObj = [self triggerWithAction:triggerDictionary[kTriggerActionKey]];
     if ([triggerObj paramIsPopupButton]) {
         triggerDictionary[kTriggerParameterKey] = [triggerObj defaultPopupParameterObject];
+    } else if ([triggerObj triggerOptionalDefaultParameterValueWithInterpolation:_interpolatedStringParameters.state == NSOnState]) {
+        triggerDictionary[kTriggerParameterKey] = [triggerObj triggerOptionalDefaultParameterValueWithInterpolation:_interpolatedStringParameters.state == NSOnState];
     }
     [self setTriggerDictionary:triggerDictionary forRow:rowIndex reloadData:YES];
 }
@@ -625,7 +646,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
         return;
     }
     NSMutableDictionary *triggerDictionary =
-        [[[self triggerDictionariesForCurrentProfile][rowIndex] mutableCopy] autorelease];
+        [[self triggerDictionariesForCurrentProfile][rowIndex] mutableCopy];
     Trigger *triggerObj = [self triggerWithAction:triggerDictionary[kTriggerActionKey]];
     id parameter = [triggerObj objectAtIndex:[sender indexOfSelectedItem]];
     if (parameter) {
@@ -650,7 +671,7 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
         return;
     }
     NSMutableDictionary *triggerDictionary =
-        [[[self triggerDictionariesForCurrentProfile][_textEditingRow] mutableCopy] autorelease];
+        [[self triggerDictionariesForCurrentProfile][_textEditingRow] mutableCopy];
     if ([textField.identifier isEqual:kRegexColumnIdentifier]) {
         triggerDictionary[kTriggerRegexKey] = [textField stringValue];
         [self setTriggerDictionary:triggerDictionary forRow:_textEditingRow reloadData:YES];
@@ -661,5 +682,10 @@ static NSString *const kBackgroundColorWellIdentifier = @"kBackgroundColorWellId
     _textEditingRow = -1;
 }
 
+- (void)profileDidChange {
+    _textEditingRow = -1;
+    [_tableView reloadData];
+}
+
 @end
 
