@@ -58,6 +58,7 @@ static NSString *const kGridSizeKey = @"Size";
         [self setSize:size];
         scrollRegionRows_ = VT100GridRangeMake(0, size_.height);
         scrollRegionCols_ = VT100GridRangeMake(0, size_.width);
+        _preferredCursorPosition = VT100GridCoordMake(-1, -1);
     }
     return self;
 }
@@ -105,6 +106,7 @@ static NSString *const kGridSizeKey = @"Size";
 
 - (void)markCharsDirty:(BOOL)dirty inRectFrom:(VT100GridCoord)from to:(VT100GridCoord)to {
     DLog(@"Mark rect from %@ to %@ dirty=%@ delegate=%@", VT100GridCoordDescription(from), VT100GridCoordDescription(to), @(dirty), delegate_);
+    assert(from.x <= to.x);
     if (!dirty) {
         allDirty_ = NO;
     }
@@ -201,14 +203,21 @@ static NSString *const kGridSizeKey = @"Size";
     self.cursorY = MIN(size_.height - 1, MAX(0, coord.y));
 }
 
-- (int)numberOfNonEmptyLines {
+- (int)numberOfNonEmptyLinesIncludingWhitespaceAsEmpty:(BOOL)includeWhitespace {
     int numberOfLinesUsed = size_.height;
-
+    NSMutableCharacterSet *allowedCharacters = [[[NSMutableCharacterSet alloc] init] autorelease];
+    [allowedCharacters addCharactersInRange:NSMakeRange(0, 1)];
+    if (includeWhitespace) {
+        [allowedCharacters addCharactersInString:@" \t"];
+        [allowedCharacters addCharactersInRange:NSMakeRange(TAB_FILLER, 1)];
+        [allowedCharacters addCharactersInRange:NSMakeRange(DWC_RIGHT, 1)];
+        [allowedCharacters addCharactersInRange:NSMakeRange(DWC_SKIP, 1)];
+    }
     for(; numberOfLinesUsed > 0; numberOfLinesUsed--) {
         screen_char_t *line = [self screenCharsAtLineNumber:numberOfLinesUsed - 1];
         int i;
         for (i = 0; i < size_.width; i++) {
-            if (line[i].code) {
+            if (line[i].complexChar || ![allowedCharacters characterIsMember:line[i].code]) {
                 break;
             }
         }
@@ -221,7 +230,7 @@ static NSString *const kGridSizeKey = @"Size";
 }
 
 - (int)numberOfLinesUsed {
-    return MAX(MIN(size_.height, cursor_.y + 1), self.numberOfNonEmptyLines);
+    return MAX(MIN(size_.height, cursor_.y + 1), [self numberOfNonEmptyLinesIncludingWhitespaceAsEmpty:NO]);
 }
 
 - (int)appendLines:(int)numLines
@@ -284,6 +293,10 @@ static NSString *const kGridSizeKey = @"Size";
     return [[self lineInfoAtLineNumber:y] timestamp];
 }
 
+- (NSInteger)generationForLine:(int)y {
+    return [[self lineInfoAtLineNumber:y] generation];
+}
+
 - (int)lengthOfLineNumber:(int)lineNumber {
     screen_char_t *line = [self screenCharsAtLineNumber:lineNumber];
     return [self lengthOfLine:line];
@@ -334,7 +347,7 @@ static NSString *const kGridSizeKey = @"Size";
 
     // Increment screenTop_, effectively scrolling the lines & dirty up by one line.
     screenTop_ = (screenTop_ + 1) % size_.height;
-
+    _haveScrolled = YES;
     // Empty contents of last line on screen.
     NSMutableData *lastLineData = [self lineDataAtLineNumber:(size_.height - 1)];
     if (lastLineData) {  // This if statement is just to quiet the analyzer.
@@ -398,12 +411,13 @@ static NSString *const kGridSizeKey = @"Size";
 
 - (int)resetWithLineBuffer:(LineBuffer *)lineBuffer
         unlimitedScrollback:(BOOL)unlimitedScrollback
-        preserveCursorLine:(BOOL)preserveCursorLine {
+        preserveCursorLine:(BOOL)preserveCursorLine
+     additionalLinesToSave:(int)additionalLinesToSave {
     self.scrollRegionRows = VT100GridRangeMake(0, size_.height);
     self.scrollRegionCols = VT100GridRangeMake(0, size_.width);
     int numLinesToScroll;
     if (preserveCursorLine) {
-        numLinesToScroll = cursor_.y;
+        numLinesToScroll = MAX(0, cursor_.y - additionalLinesToSave);
     } else {
         numLinesToScroll = [self lineNumberOfLastNonEmptyLine] + 1;
     }
@@ -416,7 +430,7 @@ static NSString *const kGridSizeKey = @"Size";
     }
     self.cursor = VT100GridCoordMake(0, 0);
 
-    [self setCharsFrom:VT100GridCoordMake(0, preserveCursorLine ? 1 : 0)
+    [self setCharsFrom:VT100GridCoordMake(0, preserveCursorLine ? 1 + additionalLinesToSave : 0)
                     to:VT100GridCoordMake(size_.width - 1, size_.height - 1)
                 toChar:[self defaultChar]];
 
@@ -437,7 +451,7 @@ static NSString *const kGridSizeKey = @"Size";
 - (int)moveCursorDownOneLineScrollingIntoLineBuffer:(LineBuffer *)lineBuffer
                                 unlimitedScrollback:(BOOL)unlimitedScrollback
                             useScrollbackWithRegion:(BOOL)useScrollbackWithRegion
-                                         willScroll:(void (^)())willScroll {
+                                         willScroll:(void (^)(void))willScroll {
     // This doesn't call -bottomMargin because it was a hotspot in profiling.
     const int scrollBottom = VT100GridRangeMax(scrollRegionRows_);
 
@@ -615,6 +629,20 @@ static NSString *const kGridSizeKey = @"Size";
     }
 }
 
+- (void)setURLCode:(unsigned short)code
+        inRectFrom:(VT100GridCoord)from
+                to:(VT100GridCoord)to {
+    for (int y = from.y; y <= to.y; y++) {
+        screen_char_t *line = [self screenCharsAtLineNumber:y];
+        for (int x = from.x; x <= to.x; x++) {
+            line[x].urlCode = code;
+        }
+        [self markCharsDirty:YES
+                  inRectFrom:VT100GridCoordMake(from.x, y)
+                          to:VT100GridCoordMake(to.x, y)];
+    }
+}
+
 - (void)copyCharsFromGrid:(VT100Grid *)otherGrid {
     if (otherGrid == self) {
         return;
@@ -882,18 +910,22 @@ static NSString *const kGridSizeKey = @"Size";
         if (aLine[cursor_.x].code == DWC_RIGHT) {
 #ifdef VERBOSE_STRING
             NSLog(@"Wiping out the right-half DWC at the cursor before writing to screen");
-            NSAssert(cursor_.x > 0, @"DWC split");  // there should never be the second half of a DWC at x=0
+            ITUpgradedNSAssert(cursor_.x > 0, @"DWC split");  // there should never be the second half of a DWC at x=0
 #endif
             aLine[cursor_.x].code = 0;
             aLine[cursor_.x].complexChar = NO;
-            aLine[cursor_.x-1].code = 0;
-            aLine[cursor_.x-1].complexChar = NO;
+            if (cursor_.x > 0) {
+                aLine[cursor_.x-1].code = 0;
+                aLine[cursor_.x-1].complexChar = NO;
+            }
             [self markCharDirty:YES
                              at:VT100GridCoordMake(cursor_.x, lineNumber)
                 updateTimestamp:YES];
-            [self markCharDirty:YES
-                             at:VT100GridCoordMake(cursor_.x - 1, lineNumber)
-                updateTimestamp:YES];
+            if (cursor_.x > 0) {
+                [self markCharDirty:YES
+                                 at:VT100GridCoordMake(cursor_.x - 1, lineNumber)
+                    updateTimestamp:YES];
+            }
         }
 
         // This is an ugly little optimization--if we're inserting just one character, see if it would
@@ -1243,7 +1275,9 @@ static NSString *const kGridSizeKey = @"Size";
                     dirtyline[ox] = ':';
                 }
             }
-            if (p[x].code && !p[x].complexChar) {
+            if (p[x].image) {
+                line[ox] = 'I';
+            } else if (p[x].code && !p[x].complexChar) {
                 if (p[x].code > 0 && p[x].code < 128) {
                     line[ox] = p[x].code;
                 } else if (p[x].code == DWC_RIGHT) {
@@ -1268,23 +1302,23 @@ static NSString *const kGridSizeKey = @"Size";
 }
 
 - (VT100GridRun)gridRunFromRange:(NSRange)range relativeToRow:(int)row {
-    NSInteger location = range.location;
-    NSInteger length = range.length;
-    if (row < 0) {
-        location += row * size_.width;
-        length += row * size_.width;
+    const NSInteger longRow = row;
+    const NSInteger location = (NSInteger)range.location + longRow * (NSInteger)size_.width;
+    const NSInteger length = range.length;
+    const NSInteger overage = MAX(0, -location);
+    const NSInteger adjustedLocation = location + overage;
+    const NSInteger adjustedLength = length - overage;
 
-        if (location < 0 || length <= 0) {
-            return VT100GridRunMake(0, 0, 0);
-        }
+    if (adjustedLocation < 0 || adjustedLength < 0) {
+        return VT100GridRunMake(0, 0, 0);
     }
 
-    return VT100GridRunMake(location % size_.width,
-                            row + location / size_.width,
-                            length);
+    return VT100GridRunMake(adjustedLocation % size_.width,
+                            adjustedLocation / size_.width,
+                            adjustedLength);
 }
 
-- (void)restoreScreenFromLineBuffer:(LineBuffer *)lineBuffer
+- (BOOL)restoreScreenFromLineBuffer:(LineBuffer *)lineBuffer
                     withDefaultChar:(screen_char_t)defaultChar
                   maxLinesToRestore:(int)maxLines {
     // Move scrollback lines into screen
@@ -1349,6 +1383,7 @@ static NSString *const kGridSizeKey = @"Size";
         }
         --destLineNumber;
     }
+    return foundCursor;
 }
 
 
@@ -1597,8 +1632,13 @@ static NSString *const kGridSizeKey = @"Size";
               kGridSizeKey: [NSDictionary dictionaryWithGridSize:size_] };
 }
 
++ (VT100GridSize)sizeInStateDictionary:(NSDictionary *)dict {
+    VT100GridSize size = [dict[kGridSizeKey] gridSize];
+    return size;
+}
+
 - (void)setStateFromDictionary:(NSDictionary *)dict {
-    if (!dict) {
+    if (!dict || [dict isKindOfClass:[NSNull class]]) {
         return;
     }
     VT100GridSize size = [dict[kGridSizeKey] gridSize];
@@ -1620,6 +1660,17 @@ static NSString *const kGridSizeKey = @"Size";
     }
 }
 
+- (void)restorePreferredCursorPositionIfPossible {
+    if (_preferredCursorPosition.x >= 0 &&
+        _preferredCursorPosition.y >= 0 &&
+        _preferredCursorPosition.x <= size_.width &&
+        _preferredCursorPosition.y < size_.height) {
+        DLog(@"Restore preferred cursor position to %@", VT100GridCoordDescription(_preferredCursorPosition));
+        self.cursor = _preferredCursorPosition;
+        _preferredCursorPosition = VT100GridCoordMake(-1, -1);
+    }
+}
+
 #pragma mark - Private
 
 - (NSMutableArray *)linesWithSize:(VT100GridSize)size {
@@ -1649,6 +1700,11 @@ static NSString *const kGridSizeKey = @"Size";
     CopyForegroundColor(&c, fg);
     CopyBackgroundColor(&c, bg);
 
+    c.underline = NO;
+    c.strikethrough = NO;
+    c.urlCode = 0;
+    c.image = 0;
+
     return c;
 }
 
@@ -1850,29 +1906,6 @@ static NSString *const kGridSizeKey = @"Size";
     return VT100GridCoordMake(cx, cy);
 }
 
-// Add a combining char to the cell at the cursor position if possible. Returns
-// YES if it is able to and NO if there is no base character to combine with.
-- (BOOL)addCombiningChar:(unichar)combiningChar toCoord:(VT100GridCoord)coord
-{
-    int cx = coord.x;
-    int cy = coord.y;
-    screen_char_t* theLine = [self screenCharsAtLineNumber:cy];
-    if (theLine[cx].code == 0 ||
-        (theLine[cx].code >= ITERM2_PRIVATE_BEGIN && theLine[cx].code <= ITERM2_PRIVATE_END) ||
-        (IsLowSurrogate(combiningChar) && !IsHighSurrogate(theLine[cx].code)) ||
-        (!IsLowSurrogate(combiningChar) && IsHighSurrogate(theLine[cx].code))) {
-        // Unable to combine.
-        return NO;
-    }
-    if (theLine[cx].complexChar) {
-        theLine[cx].code = AppendToComplexChar(theLine[cx].code,
-                                               combiningChar);
-    } else {
-        BeginComplexChar(theLine + cx, combiningChar, [delegate_ gridUseHFSPlusMapping]);
-    }
-    return YES;
-}
-
 - (NSString *)stringForCharacterAt:(VT100GridCoord)coord {
     screen_char_t *theLine = [self screenCharsAtLineNumber:coord.y];
     if (!theLine) {
@@ -1950,49 +1983,6 @@ static void DumpBuf(screen_char_t* p, int n) {
     }
 }
 
-// Find all the lines starting at startScreenY that have non-hard EOLs. Combine them into a string and return it.
-// Store the number of screen lines in *numLines
-// Store an array of UTF-16 codes in backingStorePtr, which the caller must free
-// Store an array of offsets between chars in the string and screen_char_t indices in deltasPtr, which the caller must free.
-- (NSString *)joinedLineBeginningAtLineNumber:(int)startScreenY
-                                  numLinesPtr:(int *)numLines
-                              backingStorePtr:(unichar **)backingStorePtr  // caller must free
-                                    deltasPtr:(int **)deltasPtr            // caller must free
-{
-    // Count the number of screen lines that have soft/dwc newlines beginning at
-    // line startScreenY.
-    int limitY;
-    for (limitY = startScreenY; limitY < size_.height; limitY++) {
-        screen_char_t *screenLine = [self screenCharsAtLineNumber:limitY];
-        if (screenLine[size_.width].code == EOL_HARD) {
-            break;
-        }
-    }
-    *numLines = limitY - startScreenY + 1;
-
-    // Create a single array of screen_char_t's that has those screen lines
-    // concatenated together in "temp".
-    NSMutableData *tempData = [NSMutableData dataWithLength:sizeof(screen_char_t) * size_.width * *numLines];
-    screen_char_t *temp = (screen_char_t *)[tempData mutableBytes];
-    int i = 0;
-    for (int y = startScreenY; y <= limitY; y++, i++) {
-        screen_char_t *screenLine = [self screenCharsAtLineNumber:y];
-        memcpy(temp + size_.width * i,
-               screenLine,
-               size_.width * sizeof(screen_char_t));
-    }
-
-    // Convert "temp" into an NSString. backingStorePtr and deltasPtr are filled
-    // in with malloc'ed pointers that the caller must free.
-    NSString *screenLine = ScreenCharArrayToString(temp,
-                                                   0,
-                                                   size_.width * *numLines,
-                                                   backingStorePtr,
-                                                   deltasPtr);
-
-    return screenLine;
-}
-
 // Number of cells between two coords, including both endpoints For example, with a grid of
 // xxyy
 // yyxx
