@@ -7,7 +7,7 @@
  **
  **  Project: iTerm
  **
- **  Description: overrides sendEvent: so that key mappings with command mask  
+ **  Description: overrides sendEvent: so that key mappings with command mask
  **               are handled properly.
  **
  **  This program is free software; you can redistribute it and/or modify
@@ -32,12 +32,19 @@
 #import "iTermHotKeyController.h"
 #import "iTermKeyBindingMgr.h"
 #import "iTermModifierRemapper.h"
+#import "iTermNotificationCenter.h"
 #import "iTermPreferences.h"
 #import "iTermScriptingWindow.h"
 #import "iTermShortcutInputView.h"
+#import "iTermWindowHacks.h"
 #import "NSArray+iTerm.h"
+#import "NSEvent+iTerm.h"
+#import "NSDictionary+iTerm.h"
+#import "NSResponder+iTerm.h"
 #import "NSTextField+iTerm.h"
+#import "NSView+iTerm.h"
 #import "NSWindow+iTerm.h"
+#import "NSImage+iTerm.h"
 #import "PreferencePanel.h"
 #import "PseudoTerminal.h"
 #import "PTYSession.h"
@@ -46,24 +53,104 @@
 #import "PTYWindow.h"
 
 unsigned short iTermBogusVirtualKeyCode = 0xffff;
+NSString *const iTermApplicationCharacterPaletteWillOpen = @"iTermApplicationCharacterPaletteWillOpen";
+NSString *const iTermApplicationCharacterPaletteDidClose = @"iTermApplicationCharacterPaletteDidClose";
+
+NSString *const iTermApplicationInputMethodEditorDidOpen = @"iTermApplicationInputMethodEditorDidOpen";
+NSString *const iTermApplicationInputMethodEditorDidClose = @"iTermApplicationInputMethodEditorDidClose";
+
+NSString *const iTermApplicationWillShowModalWindow = @"iTermApplicationWillShowModalWindow";
+NSString *const iTermApplicationDidCloseModalWindow = @"iTermApplicationDidCloseModalWindow";
 
 @interface iTermApplication()
 @property(nonatomic, retain) NSStatusItem *statusBarItem;
 @end
 
-@implementation iTermApplication
+static const char *iTermApplicationKVOKey = "iTermApplicationKVOKey";
+
+@interface iTermApplication()
+@property(nonatomic, retain, readwrite) NSWindow *it_windowBecomingKey;
+@end
+
+@implementation iTermApplication {
+    BOOL _it_characterPanelIsOpen;
+    BOOL _it_characterPanelShouldOpenSoon;
+}
 
 - (void)dealloc {
     [_fakeCurrentEvent release];
+    [_statusBarItem release];
+    [self removeObserver:self forKeyPath:@"modalWindow"];
+    [_it_windowBecomingKey release];
     [super dealloc];
 }
 
 + (iTermApplication *)sharedApplication {
     __kindof NSApplication *sharedApplication = [super sharedApplication];
     assert([sharedApplication isKindOfClass:[iTermApplication class]]);
+    static dispatch_once_t onceToken;
+    dispatch_once(&onceToken, ^{
+        [sharedApplication addObserver:sharedApplication
+                            forKeyPath:@"modalWindow"
+                               options:NSKeyValueObservingOptionNew | NSKeyValueObservingOptionOld
+                               context:(void *)iTermApplicationKVOKey];
+        [[NSNotificationCenter defaultCenter] addObserver:sharedApplication
+                                                 selector:@selector(it_windowDidOrderOnScreen:)
+                                                     name:@"NSWindowDidOrderOnScreenAndFinishAnimatingNotification"
+                                                   object:nil];
+        [[NSNotificationCenter defaultCenter] addObserver:sharedApplication
+                                                 selector:@selector(it_windowDidOrderOffScreen:)
+                                                     name:@"NSWindowDidOrderOffScreenAndFinishAnimatingNotification"
+                                                   object:nil];
+
+    });
     return sharedApplication;
 }
 
+// Giant pile of private API hacks for issue 7521.
+- (void)it_windowDidOrderOnScreen:(NSNotification *)notification {
+    DLog(@"windowDidOrderOnScreen");
+    NSObject *object = notification.object;
+    if ([NSStringFromClass(object.class) isEqualToString:@"NSPanelViewBridge"]) {
+        _it_imeOpen = YES;
+        [[NSNotificationCenter defaultCenter] postNotificationName:iTermApplicationInputMethodEditorDidOpen object:nil];
+    }
+}
+
+- (void)it_windowDidOrderOffScreen:(NSNotification *)notification {
+    DLog(@"windowDidOrderOffScreen");
+    NSObject *object = notification.object;
+    if ([NSStringFromClass(object.class) isEqualToString:@"NSPanelViewBridge"]) {
+        _it_imeOpen = NO;
+        [[NSNotificationCenter defaultCenter] postNotificationName:iTermApplicationInputMethodEditorDidClose object:nil];
+    }
+}
+
+- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary<NSKeyValueChangeKey,id> *)change context:(void *)context {
+    if (context != iTermApplicationKVOKey) {
+        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
+        return;
+    }
+    if ([keyPath isEqualToString:@"modalWindow"]) {
+        NSDictionary *dict = [change dictionaryByRemovingNullValues];
+        [self it_modalWindowDidChangeFrom:dict[NSKeyValueChangeOldKey] to:dict[NSKeyValueChangeNewKey]];
+        return;
+    }
+}
+
+- (void)it_modalWindowDidChangeFrom:(NSWindow *)oldValue to:(NSWindow *)newValue {
+    DLog(@"modal window did change from %@ to %@", oldValue, newValue);
+    if (oldValue == nil && newValue != nil) {
+        _it_modalWindowOpen = YES;
+        [[NSNotificationCenter defaultCenter] postNotificationName:iTermApplicationWillShowModalWindow object:nil];
+        return;
+    }
+    if (oldValue != nil && newValue == nil) {
+        _it_modalWindowOpen = NO;
+        [[NSNotificationCenter defaultCenter] postNotificationName:iTermApplicationDidCloseModalWindow object:nil];
+    }
+}
+
 - (BOOL)_eventUsesNavigationKeys:(NSEvent*)event {
     NSString* unmodkeystr = [event charactersIgnoringModifiers];
     if ([unmodkeystr length] == 0) {
@@ -100,12 +187,13 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
 
 - (BOOL)routeEventToShortcutInputView:(NSEvent *)event {
     NSResponder *firstResponder = [[NSApp keyWindow] firstResponder];
-    if (event.keyCode == iTermBogusVirtualKeyCode) {
-        return YES;
-    }
     if ([firstResponder isKindOfClass:[iTermShortcutInputView class]]) {
         iTermShortcutInputView *shortcutView = (iTermShortcutInputView *)firstResponder;
         if (shortcutView) {
+            if (event.keyCode == iTermBogusVirtualKeyCode) {
+                // You can't register a carbon hotkey for these so just ignore them when listening for a shortcut.
+                return YES;
+            }
             [shortcutView handleShortcutEvent:event];
             return YES;
         }
@@ -114,7 +202,7 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
 }
 
 - (int)digitKeyForEvent:(NSEvent *)event {
-    if ([iTermAdvancedSettingsModel useVirtualKeyCodesForDetectingDigits]) {
+    if ([iTermPreferences boolForKey:kPreferenceKeyEmulateUSKeyboard]) {
         switch (event.keyCode) {
             case kVK_ANSI_1:
             case kVK_ANSI_Keypad1:
@@ -156,8 +244,10 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
 
 - (BOOL)switchToWindowByNumber:(NSEvent *)event {
     const NSUInteger allModifiers =
-        (NSShiftKeyMask | NSControlKeyMask | NSCommandKeyMask | NSAlternateKeyMask);
-    if (([event modifierFlags] & allModifiers) == [iTermPreferences maskForModifierTag:[iTermPreferences intForKey:kPreferenceKeySwitchWindowModifier]]) {
+        (NSEventModifierFlagShift | NSEventModifierFlagControl | NSEventModifierFlagCommand | NSEventModifierFlagOption);
+    const NSUInteger maskedValue = ([event it_modifierFlags] & allModifiers);
+    const NSUInteger requirement = [iTermPreferences maskForModifierTag:[iTermPreferences intForKey:kPreferenceKeySwitchWindowModifier]];
+    if (maskedValue == requirement) {
         // Command-Alt (or selected modifier) + number: Switch to window by number.
         int digit = [self digitKeyForEvent:event];
         if (digit >= 1 && digit <= 9) {
@@ -179,8 +269,8 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
 }
 
 - (BOOL)switchToPaneInWindowController:(PseudoTerminal *)currentTerminal byNumber:(NSEvent *)event {
-    const int mask = NSShiftKeyMask | NSControlKeyMask | NSAlternateKeyMask | NSCommandKeyMask;
-    if (([event modifierFlags] & mask) == [iTermPreferences maskForModifierTag:[iTermPreferences intForKey:kPreferenceKeySwitchPaneModifier]]) {
+    const int mask = NSEventModifierFlagShift | NSEventModifierFlagControl | NSEventModifierFlagOption | NSEventModifierFlagCommand;
+    if (([event it_modifierFlags] & mask) == [iTermPreferences maskForModifierTag:[iTermPreferences intForKey:kPreferenceKeySwitchPaneModifier]]) {
         int digit = [self digitKeyForEvent:event];
         NSArray *orderedSessions = currentTerminal.currentTab.orderedSessions;
         int numSessions = [orderedSessions count];
@@ -196,13 +286,17 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
             [currentTerminal.currentTab setActiveSession:orderedSessions[digit - 1]];
             return YES;
         }
+        if (digit >= 1 && digit <= 9) {
+            // Ignore Modifier+Number if there's no matching split pane. Issue 6624.
+            return YES;
+        }
     }
     return NO;
 }
 
 - (BOOL)switchToTabInTabView:(PTYTabView *)tabView byNumber:(NSEvent *)event {
-    const int mask = NSShiftKeyMask | NSControlKeyMask | NSAlternateKeyMask | NSCommandKeyMask;
-    if (([event modifierFlags] & mask) == [iTermPreferences maskForModifierTag:[iTermPreferences intForKey:kPreferenceKeySwitchTabModifier]]) {
+    const int mask = NSEventModifierFlagShift | NSEventModifierFlagControl | NSEventModifierFlagOption | NSEventModifierFlagCommand;
+    if (([event it_modifierFlags] & mask) == [iTermPreferences maskForModifierTag:[iTermPreferences intForKey:kPreferenceKeySwitchTabModifier]]) {
         int digit = [self digitKeyForEvent:event];
         if (digit == 9 && [tabView numberOfTabViewItems] > 0) {
             // Command (or selected modifier)+9: Switch to last tab if there are fewer than 9.
@@ -216,6 +310,10 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
             [tabView selectTabViewItemAtIndex:digit-1];
             return YES;
         }
+        if (digit >= 1 && digit <= 9) {
+            // Ignore Modifier+Number if there's no matching tab. Issue 6624.
+            return YES;
+        }
     }
     return NO;
 }
@@ -232,7 +330,10 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
         }
     }
 
-    if (okToRemap && [currentSession hasActionableKeyMappingForEvent:event] && !currentSession.copyMode) {
+    if (okToRemap && [currentSession hasActionableKeyMappingForEvent:event]) {
+        if (currentSession.copyMode && [currentSession copyModeConsumesEvent:event]) {
+            return NO;
+        }
         // Remap key.
         DLog(@"Remapping to actionable event");
         [currentSession keyDown:event];
@@ -301,7 +402,7 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
     if ([self routeEventToShortcutInputView:event]) {
         return YES;
     }
-
+    [[iTermFlagsChangedNotification notificationWithEvent:event] post];
     return NO;
 }
 
@@ -331,20 +432,43 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
     return NO;
 }
 
+- (void)handleScrollWheelEvent:(NSEvent *)event {
+    NSPoint point = event.locationInWindow;
+    if (event.window) {
+        point = [event.window convertPointToScreen:point];
+    }
+    NSView *current = [NSView viewAtScreenCoordinate:point];
+    if (current.window != event.window) {
+        return;
+    }
+    while (current) {
+        if ([current respondsToSelector:@selector(it_wantsScrollWheelMomentumEvents)] &&
+            [current it_wantsScrollWheelMomentumEvents]) {
+            [current it_scrollWheelMomentum:event];
+            return;
+        }
+        current = current.superview;
+    }
+}
+
 // override to catch key press events very early on
 - (void)sendEvent:(NSEvent *)event {
-    if ([event type] == NSFlagsChanged) {
+    if ([event type] == NSEventTypeFlagsChanged) {
         event = [self eventByRemappingForSecureInput:event];
         if ([self handleFlagsChangedEvent:event]) {
             return;
         }
-    } else if ([event type] == NSKeyDown) {
+    } else if ([event type] == NSEventTypeKeyDown) {
         event = [self eventByRemappingForSecureInput:event];
         if ([self handleKeyDownEvent:event]) {
             return;
         }
         DLog(@"NSKeyDown event taking the regular path");
+    } else if (event.type == NSEventTypeScrollWheel && (event.momentumPhase == NSEventPhaseChanged ||
+                                                        event.momentumPhase == NSEventPhaseEnded)) {
+        [self handleScrollWheelEvent:event];
     }
+
     [super sendEvent:event];
 }
 
@@ -388,11 +512,11 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
         });
 
         if ([iTermAdvancedSettingsModel statusBarIcon]) {
-            NSImage *image = [NSImage imageNamed:@"StatusItem"];
-            self.statusBarItem = [[[NSStatusBar systemStatusBar] statusItemWithLength:image.size.width] retain];
+            NSImage *image = [NSImage it_imageNamed:@"StatusItem" forClass:self.class];
+            self.statusBarItem = [[NSStatusBar systemStatusBar] statusItemWithLength:image.size.width];
             _statusBarItem.title = @"";
             _statusBarItem.image = image;
-            _statusBarItem.alternateImage = [NSImage imageNamed:@"StatusItemAlt"];
+            _statusBarItem.alternateImage = [NSImage it_imageNamed:@"StatusItemAlt" forClass:self.class];
             _statusBarItem.highlightMode = YES;
 
             _statusBarItem.menu = [[self delegate] statusBarMenu];
@@ -430,9 +554,44 @@ unsigned short iTermBogusVirtualKeyCode = 0xffff;
                                                                      DLog(@"Application did become active completion block finished. Removing observer.");
                                                                      [[NSNotificationCenter defaultCenter] removeObserver:observer];
                                                                  }];
-        [self activateIgnoringOtherApps:YES];
+        // It's not clear how this differs from [self activateIgnoringOtherApps:YES], but on 10.13
+        // it does not cause previously ordered-out windows to be ordered over other applications'
+        // windows. See issue 6875.
+        [[NSRunningApplication currentApplication] activateWithOptions:NSApplicationActivateIgnoringOtherApps];
     }
 }
 
+- (BOOL)it_characterPanelIsOpen {
+    return _it_characterPanelShouldOpenSoon || _it_characterPanelIsOpen;
+}
+
+- (void)orderFrontCharacterPalette:(id)sender {
+    _it_characterPanelShouldOpenSoon = YES;
+    [[NSNotificationCenter defaultCenter] postNotificationName:iTermApplicationCharacterPaletteWillOpen
+                                                        object:nil];
+    [super orderFrontCharacterPalette:sender];
+    const NSTimeInterval deadlineToOpen = ([NSDate timeIntervalSinceReferenceDate] +
+                                           [iTermAdvancedSettingsModel timeToWaitForEmojiPanel]);
+    [iTermWindowHacks pollForCharacterPanelToOpenOrCloseWithCompletion:^BOOL(BOOL open) {
+        if (open && _it_characterPanelShouldOpenSoon) {
+            _it_characterPanelShouldOpenSoon = NO;
+            _it_characterPanelIsOpen = YES;
+        } else if (!open && self.it_characterPanelIsOpen) {
+            self->_it_characterPanelShouldOpenSoon = NO;
+            self->_it_characterPanelIsOpen = NO;
+            [[NSNotificationCenter defaultCenter] postNotificationName:iTermApplicationCharacterPaletteDidClose
+                                                                object:nil];
+        }
+        return open || ([NSDate timeIntervalSinceReferenceDate] < deadlineToOpen);  // keep running while open
+    }];
+}
+
+- (void)it_makeWindowKey:(NSWindow *)window {
+    NSWindow *saved = [self.it_windowBecomingKey retain];
+    self.it_windowBecomingKey = window;
+    [window makeKeyAndOrderFront:nil];
+    self.it_windowBecomingKey = [saved autorelease];
+}
+
 @end
 
