@@ -3,7 +3,7 @@
 @class iTermLogoGenerator;
 @class iTermOpenQuicklyTableCellView;
 
-// Represents and item in the Open Quicky table.
+// Represents and item in the Open Quickly table.
 @interface iTermOpenQuicklyItem : NSObject
 
 // Globally unique identifier for represented object.
@@ -43,7 +43,23 @@
 @end
 
 @interface iTermOpenQuicklyArrangementItem : iTermOpenQuicklyItem
+@property (nonatomic) BOOL inTabs;
 @end
 
 @interface iTermOpenQuicklyHelpItem : iTermOpenQuicklyItem
 @end
+
+@interface iTermOpenQuicklyScriptItem : iTermOpenQuicklyItem
+@end
+
+@interface iTermOpenQuicklyColorPresetItem : iTermOpenQuicklyItem
+// Holds the session's colors and can create a logo with them as needed.
+@property(nonatomic, retain) iTermLogoGenerator *logoGenerator;
+@property(nonatomic, copy) NSString *presetName;
+@end
+
+@class iTermAction;
+@interface iTermOpenQuicklyActionItem : iTermOpenQuicklyItem
+@property(nonatomic, strong) iTermAction *action;
+@end
+
