@@ -3,11 +3,13 @@
 #import "iTermPopupWindowController.h"
 #import "DebugLogging.h"
 #import "iTermAdvancedSettingsModel.h"
+#import "NSTextField+iTerm.h"
 #import "NSWindow+PSM.h"
 #import "PopupEntry.h"
 #import "PopupModel.h"
 #import "PopupWindow.h"
 #import "PTYTextView.h"
+#import "SolidColorView.h"
 #import "VT100Screen.h"
 
 #import <QuartzCore/QuartzCore.h>
@@ -18,24 +20,24 @@
 @implementation iTermPopupWindowController {
     // Subclass-owned tableview.
     NSTableView* tableView_;
-    
+
     // Results currently being displayed.
     PopupModel* model_;
-    
-    // All candidate results, including those not matching filter. Subclass-owend.
+
+    // All candidate results, including those not matching filter. Subclass-owned.
     PopupModel* unfilteredModel_;
-    
+
     // Timer to set clearFilterOnNextKeyDown_.
     NSTimer* timer_;
-    
+
     // If set, then next time a key is pressed erase substring_ before appending.
     BOOL clearFilterOnNextKeyDown_;
     // What the user has typed so far to filter result set.
     NSMutableString* substring_;
-    
+
     // If true then window is above cursor.
     BOOL onTop_;
-    
+
     // Set to true when the user changes the selected row.
     BOOL haveChangedSelection_;
     // String that the user has selected.
@@ -43,6 +45,9 @@
 
     // True while reloading data.
     BOOL reloading_;
+
+    NSString *footerString_;
+    NSTextField *footerView_;
 }
 
 - (instancetype)initWithWindowNibName:(NSString*)nibName tablePtr:(NSTableView**)table model:(PopupModel*)model {
@@ -70,9 +75,39 @@
     [substring_ release];
     [model_ release];
     [tableView_ release];
+    [footerString_ release];
+    [footerView_ release];
     [super dealloc];
 }
 
+- (void)awakeFromNib {
+    [super awakeFromNib];
+    footerString_ = [[self footerString] copy];
+    if (!footerString_) {
+        return;
+    }
+
+    footerView_ = [NSTextField newLabelStyledTextField];
+    [self updateBackgroundColor];
+    footerView_.stringValue = footerString_;
+    [self.window.contentView addSubview:footerView_];
+    [footerView_ sizeToFit];
+    const CGFloat footerHeight = footerView_.frame.size.height;
+    footerView_.frame = NSMakeRect(0, 0, self.window.contentView.frame.size.width - footerHeight, footerHeight);
+    footerView_.autoresizingMask = (NSViewWidthSizable | NSViewMaxYMargin);
+
+    NSRect frame = [[tableView_ enclosingScrollView] frame];
+    frame.size.height = MAX(0, frame.size.height - footerHeight);
+    frame.origin.y += footerHeight;
+    [[tableView_ enclosingScrollView] setFrame:frame];
+}
+
+- (void)updateBackgroundColor {
+    if (footerString_) {
+        self.window.backgroundColor = [NSColor whiteColor];
+    }
+}
+
 - (void)shutdown
 {
     // Disable the fancy footwork in -[PopupWindow close]
@@ -88,6 +123,7 @@
 - (void)setTableView:(NSTableView *)table {
     [tableView_ autorelease];
     tableView_ = [table retain];
+    [self updateBackgroundColor];
 }
 
 - (BOOL)disableFocusFollowsMouse
@@ -95,26 +131,79 @@
     return YES;
 }
 
+- (PopupWindow *)popupWindow {
+    return (PopupWindow *)self.window;
+}
+
 - (void)popWithDelegate:(id<PopupDelegate>)delegate {
     self.delegate = delegate;
 
-    [[self window] setParentWindow:delegate.popupWindowController.window];
-    if (delegate.popupWindowIsInHotkeyWindow &&
-        [iTermAdvancedSettingsModel hotkeyWindowFloatsAboveOtherWindows]) {
-        self.window.level = NSPopUpMenuWindowLevel;
-    }
+    [self.popupWindow setOwningWindow:delegate.popupWindowController.window];
 
     static const NSTimeInterval kAnimationDuration = 0.15;
     self.window.alphaValue = 0;
-    [self showWindow:delegate.popupWindowController];
-    [[self window] makeKeyAndOrderFront:delegate.popupWindowController];
-
+    if (delegate.popupWindowIsInFloatingHotkeyWindow) {
+        self.window.styleMask = NSWindowStyleMaskNonactivatingPanel;
+        self.window.collectionBehavior = (NSWindowCollectionBehaviorCanJoinAllSpaces |
+                                          NSWindowCollectionBehaviorIgnoresCycle |
+                                          NSWindowCollectionBehaviorFullScreenAuxiliary);
+        self.window.level = NSPopUpMenuWindowLevel;
+        [self.window makeKeyAndOrderFront:nil];
+    } else {
+        [self showWindow:delegate.popupWindowController];
+        [[self window] makeKeyAndOrderFront:delegate.popupWindowController];
+    }
     [NSAnimationContext beginGrouping];
     [[NSAnimationContext currentContext] setDuration:kAnimationDuration];
     self.window.animator.alphaValue = 1;
     [NSAnimationContext endGrouping];
 }
 
+#pragma mark - NSResponder Hacks
+
+// When a popup window is open we will forward some actions to the "real" responder in the owning
+// window. We have to implement the methods in question so that the SDK will enable them in the
+// menu.
+- (BOOL)validateMenuItem:(NSMenuItem *)menuItem {
+    for (NSResponder *responder in self.allResponders) {
+        if ([responder respondsToSelector:@selector(validateMenuItem:)] &&
+            [responder validateMenuItem:menuItem]) {
+            return YES;
+        }
+    }
+    return NO;
+}
+
+- (NSArray *)allResponders {
+    NSMutableArray *responders = [NSMutableArray array];
+    NSResponder *responder = [self.popupWindow.owningWindow firstResponder];
+    while (responder) {
+        [responders addObject:responder];
+        responder = [responder nextResponder];
+    }
+    return responders;
+}
+
+- (void)paste:(id)sender {
+    [self.popupWindow.owningWindow.firstResponder tryToPerform:_cmd with:sender];
+}
+
+- (void)copy:(id)sender {
+    [self.popupWindow.owningWindow.firstResponder tryToPerform:_cmd with:sender];
+}
+
+- (void)copyWithStyles:(id)sender {
+    [self.popupWindow.owningWindow.firstResponder tryToPerform:_cmd with:sender];
+}
+
+- (void)selectAll:(id)sender {
+    [self.popupWindow.owningWindow.firstResponder tryToPerform:_cmd with:sender];
+}
+
+- (void)selectOutputOfLastCommand:(id)sender {
+    [self.popupWindow.owningWindow.firstResponder tryToPerform:_cmd with:sender];
+}
+
 - (PopupModel*)unfilteredModel
 {
     return unfilteredModel_;
@@ -181,6 +270,14 @@
     timer_ = nil;
 }
 
+- (CGFloat)desiredHeight {
+    CGFloat height = [[tableView_ headerView] frame].size.height + MIN(20, [model_ count]) * ([tableView_ rowHeight] + [tableView_ intercellSpacing].height);
+    if (footerView_) {
+        height += footerView_.frame.size.height;
+    }
+    return height;
+}
+
 - (void)setPosition:(BOOL)canChangeSide
 {
     BOOL onTop = NO;
@@ -192,9 +289,9 @@
     PTYTextView* tv = [self.delegate popupVT100TextView];
     [tv scrollEnd];
     NSRect frame = [[self window] frame];
-    frame.size.height = [[tableView_ headerView] frame].size.height + MIN(20, [model_ count]) * ([tableView_ rowHeight] + [tableView_ intercellSpacing].height);
+    frame.size.height = self.desiredHeight;
 
-    NSPoint p = NSMakePoint(MARGIN + cx * [tv charWidth],
+    NSPoint p = NSMakePoint([iTermAdvancedSettingsModel terminalMargin] + cx * [tv charWidth],
                             ([screen numberOfLines] - [screen height] + cy) * [tv lineHeight]);
     p = [tv convertPoint:p toView:nil];
     p = [[tv window] pointToScreenCoords:p];
@@ -223,6 +320,15 @@
 
     frame.origin = p;
     [[self window] setFrame:frame display:NO];
+
+    if (footerView_) {
+        NSRect frame = [self.window.contentView bounds];
+        const CGFloat footerHeight = footerView_.frame.size.height;
+        frame.size.height = MAX(0, frame.size.height - footerHeight);
+        frame.origin.y += footerHeight;
+        tableView_.enclosingScrollView.frame = frame;
+    }
+
     if (canChangeSide) {
         BOOL flip = (onTop != onTop_);
         [self setOnTop:onTop];
@@ -364,11 +470,11 @@
 
     NSColor* textColor;
     if (isSelected) {
-        textColor = [NSColor whiteColor];
+        textColor = [NSColor selectedMenuItemTextColor];
     } else {
-        textColor = [NSColor blackColor];
+        textColor = [NSColor labelColor];
     }
-    NSColor* lightColor = [textColor colorWithAlphaComponent:0.4];
+    NSColor* lightColor = [textColor colorWithAlphaComponent:0.7];
     NSDictionary* lightAttributes = @{ NSFontAttributeName: sysFont,
                                        NSForegroundColorAttributeName: lightColor,
                                        NSParagraphStyleAttributeName: paragraphStyle };
@@ -488,6 +594,10 @@
 {
 }
 
+- (NSString *)footerString {
+    return NULL;
+}
+
 // DataSource methods
 - (NSInteger)numberOfRowsInTableView:(NSTableView *)aTableView
 {
