@@ -8,7 +8,8 @@
 
 #import <Foundation/Foundation.h>
 #import "PreferenceInfo.h"
-#import "PSMTabBarControl.h"
+
+extern NSString *const iTermMetalSettingsDidChangeNotification;
 
 // Values for kPreferenceKeyOpenTmuxWindowsIn (corresponds to tags in control).
 typedef NS_ENUM(NSInteger, iTermOpenTmuxWindowsMode) {
@@ -22,7 +23,15 @@ typedef NS_ENUM(int, iTermPreferencesTabStyle) {
     TAB_STYLE_LIGHT = 0,
     TAB_STYLE_DARK = 1,
     TAB_STYLE_LIGHT_HIGH_CONTRAST = 2,
-    TAB_STYLE_DARK_HIGH_CONTRAST = 3
+    TAB_STYLE_DARK_HIGH_CONTRAST = 3,
+    TAB_STYLE_AUTOMATIC = 4,  // automatic + normal windows. This has the side-effect of changing compact windows to normal.
+    TAB_STYLE_MINIMAL = 5,  // minimal + compact windows. Changes normal windows to compact.
+    TAB_STYLE_COMPACT = 6  // automatic + compact windows. This should be like automatic, except it has a side-effect of changing normal windows to compact.
+};
+
+typedef NS_ENUM(NSUInteger, iTermStatusBarPosition) {
+    iTermStatusBarPositionTop,
+    iTermStatusBarPositionBottom
 };
 
 // Values for kPreferenceKeyTabPosition (corresponds to tags in control).
@@ -51,6 +60,7 @@ extern NSString *const kPreferenceKeyOpenNoWindowsAtStartup;
 extern NSString *const kPreferenceKeyQuitWhenAllWindowsClosed;
 extern NSString *const kPreferenceKeyConfirmClosingMultipleTabs;
 extern NSString *const kPreferenceKeyPromptOnQuit;
+extern NSString *const kPreferenceKeyPromptOnQuitEvenIfThereAreNoWindows;
 extern NSString *const kPreferenceKeyInstantReplayMemoryMegabytes;
 extern NSString *const kPreferenceKeySavePasteAndCommandHistory;
 extern NSString *const kPreferenceKeyAddBonjourHostsToProfiles;
@@ -67,27 +77,40 @@ extern NSString *const kPreferenceKeyCharactersConsideredPartOfAWordForSelection
 extern NSString *const kPreferenceKeySmartWindowPlacement;
 extern NSString *const kPreferenceKeyAdjustWindowForFontSizeChange;
 extern NSString *const kPreferenceKeyMaximizeVerticallyOnly;
-extern NSString *const kPreferenceKeyLionStyleFullscren;
+extern NSString *const kPreferenceKeyLionStyleFullscreen;
+extern NSString *const kPreferenceKeySeparateWindowTitlePerTab;
 extern NSString *const kPreferenceKeyOpenTmuxWindowsIn;
 extern NSString *const kPreferenceKeyTmuxDashboardLimit;
 extern NSString *const kPreferenceKeyAutoHideTmuxClientSession;
+extern NSString *const kPreferenceKeyUseTmuxProfile;
+extern NSString *const kPreferenceKeyUseTmuxStatusBar;
+extern NSString *const kPreferenceKeyUseMetal;
+extern NSString *const kPreferenceKeyDisableMetalWhenUnplugged;
+extern NSString *const kPreferenceKeyPreferIntegratedGPU;
+extern NSString *const kPreferenceKeyMetalMaximizeThroughput;
+extern NSString *const kPreferenceKeyEnableAPIServer;
 
 // Appearance
+extern NSString *const kPreferenceKeyTabStyle_Deprecated;
 extern NSString *const kPreferenceKeyTabStyle;
 extern NSString *const kPreferenceKeyTabPosition;
+extern NSString *const kPreferenceKeyStatusBarPosition;
 extern NSString *const kPreferenceKeyHideTabBar;
 extern NSString *const kPreferenceKeyHideTabNumber;
-extern NSString *const kPreferenceKeyHideTabCloseButton;
+extern NSString *const kPreferenceKeyPreserveWindowSizeWhenTabBarVisibilityChanges;
+extern NSString *const kPreferenceKeyHideTabCloseButton;  // DEPRECATED
+extern NSString *const kPreferenceKeyTabsHaveCloseButton;
 extern NSString *const kPreferenceKeyHideTabActivityIndicator;
 extern NSString *const kPreferenceKeyShowNewOutputIndicator;
 extern NSString *const kPreferenceKeyShowPaneTitles;
+extern NSString *const kPreferenceKeyPerPaneBackgroundImage;
 extern NSString *const kPreferenceKeyHideMenuBarInFullscreen;
 extern NSString *const kPreferenceKeyUIElement;
 extern NSString *const kPreferenceKeyFlashTabBarInFullscreen;
 extern NSString *const kPreferenceKeyStretchTabsToFillBar;
 extern NSString *const kPreferenceKeyShowWindowNumber;
-extern NSString *const kPreferenceKeyShowJobName;
-extern NSString *const kPreferenceKeyShowProfileName;
+extern NSString *const kPreferenceKeyShowJobName_Deprecated;  // DEPRECATED
+extern NSString *const kPreferenceKeyShowProfileName_Deprecated;  // DEPRECATED
 extern NSString *const kPreferenceKeyDimOnlyText;
 extern NSString *const kPreferenceKeyDimmingAmount;
 extern NSString *const kPreferenceKeyDimInactiveSplitPanes;
@@ -97,6 +120,7 @@ extern NSString *const kPreferenceKeyDisableFullscreenTransparencyByDefault;
 extern NSString *const kPreferenceKeyEnableDivisionView;
 extern NSString *const kPreferenceKeyEnableProxyIcon;
 extern NSString *const kPreferenceKeyDimBackgroundWindows;
+extern NSString *const kPreferenceKeySeparateStatusBarsPerPane;
 
 // Keys
 extern NSString *const kPreferenceKeyControlRemapping;
@@ -107,11 +131,15 @@ extern NSString *const kPreferenceKeyRightCommandRemapping;
 extern NSString *const kPreferenceKeySwitchPaneModifier;
 extern NSString *const kPreferenceKeySwitchTabModifier;
 extern NSString *const kPreferenceKeySwitchWindowModifier;
+extern NSString *const kPreferenceKeyEmulateUSKeyboard;  // See issue 6994
 
 extern NSString *const kPreferenceKeyHotkeyEnabled;
 extern NSString *const kPreferenceKeyHotKeyCode;
 extern NSString *const kPreferenceKeyHotkeyCharacter;
 extern NSString *const kPreferenceKeyHotkeyModifiers;
+extern NSString *const kPreferenceKeyEnableHapticFeedbackForEsc;
+extern NSString *const kPreferenceKeyEnableSoundForEsc;
+extern NSString *const kPreferenceKeyVisualIndicatorForEsc;
 
 // Migration to multi-hotkey window will move these settings into a profile.
 extern NSString *const kPreferenceKeyHotKeyTogglesWindow_Deprecated;  // Deprecated
@@ -130,6 +158,7 @@ extern NSString *const kPreferenceKeyDoubleClickPerformsSmartSelection;
 // Not in prefs
 // Stores the last CFBundleVersion run.
 extern NSString *const kPreferenceKeyAppVersion;
+extern NSString *const kPreferenceKeyAllAppVersions;
 
 // Auto-command history (set through menu)
 extern NSString *const kPreferenceAutoCommandHistory;
@@ -152,6 +181,7 @@ extern NSString *const kPreferenceKeyPasteWarningNumberOfSpacesPerTab;
 
 extern NSString *const kPreferenceKeyShowFullscreenTabBar;
 extern NSString *const kPreferenceKeyDefaultToolbeltWidth;
+extern NSString *const kPreferenceKeySizeChangesAffectProfile;
 
 // Set to YES on the first launch of a version that supports multiple hotkey windows.
 extern NSString *const kPreferenceKeyHotkeyMigratedFromSingleToMulti;
@@ -164,6 +194,7 @@ extern NSString *const kPreferenceKeyHotkeyMigratedFromSingleToMulti;
 
 // Last app version launched, if any.
 + (NSString *)appVersionBeforeThisLaunch;
++ (NSSet<NSString *> *)allAppVersionsUsedOnThisMachine;
 
 + (void)setObject:(id)object forKey:(NSString *)key;
 + (NSObject *)objectForKey:(NSString *)key;
@@ -201,3 +232,9 @@ extern NSString *const kPreferenceKeyHotkeyMigratedFromSingleToMulti;
 + (NSUInteger)maskForModifierTag:(iTermPreferencesModifierTag)tag;
 
 @end
+
+@interface iTermPreferences (FastAccessors)
++ (BOOL)hideTabActivityIndicator;
++ (BOOL)maximizeMetalThroughput;
++ (BOOL)useTmuxProfile;
+@end
