@@ -4,13 +4,15 @@
 @end
 #endif
 
+@class iTermThemeFrame;
+
 NS_ASSUME_NONNULL_BEGIN
 
 @implementation THE_CLASS
 {
     int blurFilter;
     double blurRadius_;
-    
+
     // If set, then windowWillShowInitial is not invoked.
     BOOL _layoutDone;
 
@@ -26,8 +28,12 @@ NS_ASSUME_NONNULL_BEGIN
     double _cachedTotalOcclusion;
 
     NSTimeInterval _timeOfLastWindowTitleChange;
+    BOOL _needsInvalidateShadow;
 }
 
+@synthesize it_openingSheet;
+@synthesize it_becomingKey;
+
 - (instancetype)initWithContentRect:(NSRect)contentRect
                           styleMask:(NSWindowStyleMask)aStyle
                             backing:(NSBackingStoreType)bufferingType
@@ -49,12 +55,20 @@ ITERM_WEAKLY_REFERENCEABLE
 
 - (void)iterm_dealloc {
     DLog(@"Invalidate cached occlusion: %@ %p", NSStringFromSelector(_cmd), self);
-    [[iTermWindowOcclusionChangeMonitor sharedInstance] invalidateCachedOcclusion];
+    // Not safe to call this from dealloc because can very indirectly try to retain this object.
+    dispatch_async(dispatch_get_main_queue(), ^{
+        [[iTermWindowOcclusionChangeMonitor sharedInstance] invalidateCachedOcclusion];
+    });
     [restoreState_ release];
     [super dealloc];
 
 }
 
+- (void)setDocumentEdited:(BOOL)documentEdited {
+    [super setDocumentEdited:documentEdited];
+    [[NSNotificationCenter defaultCenter] postNotificationName:iTermWindowDocumentedEditedDidChange object:self];
+}
+
 - (BOOL)titleChangedRecently {
     NSTimeInterval now = [NSDate timeIntervalSinceReferenceDate];
     return (now > _timeOfLastWindowTitleChange && now - _timeOfLastWindowTitleChange < iTermWindowTitleChangeMinimumInterval);
@@ -67,8 +81,13 @@ ITERM_WEAKLY_REFERENCEABLE
 
 - (BOOL)validateMenuItem:(NSMenuItem *)item {
     if (item.action == @selector(performMiniaturize:)) {
-        // This makes borderless windows miniaturizable.
-        return ![self.ptyDelegate anyFullScreen];
+        if (@available(macOS 10.13, *)) {
+            // Can miniaturize borderless windows
+            return ![self.ptyDelegate lionFullScreen];
+        } else {
+            // This was originally for #4402. Not sure how it worked, but I don't want to mess with 10.12.
+            return ![self.ptyDelegate anyFullScreen];
+        }
     } else {
         return [super validateMenuItem:item];
     }
@@ -76,9 +95,13 @@ ITERM_WEAKLY_REFERENCEABLE
 
 - (void)performMiniaturize:(nullable id)sender {
     if ([self.ptyDelegate anyFullScreen]) {
-        [super performMiniaturize:sender];
+        if (@available(macOS 10.13, *)) {
+            [super miniaturize:sender];
+        } else {
+            [super performMiniaturize:sender];
+        }
     } else {
-        // NSWindow's performMiniaturize gates miniaturization on the presence of a miniaturize button.
+        DLog(@"performMiniaturize calling [self miniaturize:]");
         [self miniaturize:self];
     }
 }
@@ -155,7 +178,7 @@ ITERM_WEAKLY_REFERENCEABLE
 
 - (void)toggleFullScreen:(nullable id)sender {
     if (![[self ptyDelegate] lionFullScreen]  &&
-        ![iTermPreferences boolForKey:kPreferenceKeyLionStyleFullscren]) {
+        ![iTermPreferences boolForKey:kPreferenceKeyLionStyleFullscreen]) {
         // The user must have clicked on the toolbar arrow, but the pref is set
         // to use traditional fullscreen.
         [(id<PTYWindowDelegateProtocol>)[self delegate] toggleTraditionalFullScreenMode];
@@ -200,7 +223,7 @@ ITERM_WEAKLY_REFERENCEABLE
     NSArray<NSValue *> *frames = [windows mapWithBlock:^id(NSWindow *window) {
         return [NSValue valueWithRect:window.frame];
     }];
-    
+
     double lowestCost = INFINITY;
     NSRect bestFrame = self.frame;
     const CGFloat widthToScan = screenRect.size.width - self.frame.size.width;
@@ -220,7 +243,7 @@ ITERM_WEAKLY_REFERENCEABLE
             CGFloat distanceFromScreenCenter = sqrt(pow(proposedCenter.x - screenCenter.x, 2) +
                                                     pow(proposedCenter.y - screenCenter.y, 2)) / maxDistance;
             const CGFloat cost = [self sumOfIntersectingAreaOfRect:proposedRect withRects:frames] + distanceFromScreenCenter;
-            
+
             if (cost < lowestCost) {
                 lowestCost = cost;
                 bestFrame = proposedRect;
@@ -239,6 +262,7 @@ ITERM_WEAKLY_REFERENCEABLE
 
 - (void)makeKeyAndOrderFront:(nullable id)sender {
     DLog(@"%@ makeKeyAndOrderFront: layoutDone=%@ %@", NSStringFromClass([self class]), @(_layoutDone), [NSThread callStackSymbols]);
+    DLog(@"%@\n%@", NSStringFromSelector(_cmd), [NSThread callStackSymbols]);
     if (!_layoutDone) {
         DLog(@"try to call windowWillShowInitial");
         [self setLayoutDone];
@@ -252,7 +276,9 @@ ITERM_WEAKLY_REFERENCEABLE
     DLog(@"The current window frame is %fx%f", [self frame].size.width, [self frame].size.height);
     DLog(@"Invalidate cached occlusion: %@ %p", NSStringFromSelector(_cmd), self);
     [[iTermWindowOcclusionChangeMonitor sharedInstance] invalidateCachedOcclusion];
+    self.it_becomingKey = YES;
     [super makeKeyAndOrderFront:sender];
+    self.it_becomingKey = NO;
 }
 
 - (BOOL)canBecomeKeyWindow {
@@ -271,18 +297,21 @@ ITERM_WEAKLY_REFERENCEABLE
 
 - (void)orderFrontRegardless {
     DLog(@"Invalidate cached occlusion: %@ %p", NSStringFromSelector(_cmd), self);
+    DLog(@"%@\n%@", NSStringFromSelector(_cmd), [NSThread callStackSymbols]);
     [[iTermWindowOcclusionChangeMonitor sharedInstance] invalidateCachedOcclusion];
     [super orderFrontRegardless];
 }
 
 - (void)orderFront:(nullable id)sender {
     DLog(@"Invalidate cached occlusion: %@ %p", NSStringFromSelector(_cmd), self);
+    DLog(@"%@\n%@", NSStringFromSelector(_cmd), [NSThread callStackSymbols]);
     [[iTermWindowOcclusionChangeMonitor sharedInstance] invalidateCachedOcclusion];
     [super orderFront:sender];
 }
 
 - (void)orderBack:(nullable id)sender {
     DLog(@"Invalidate cached occlusion: %@ %p", NSStringFromSelector(_cmd), self);
+    DLog(@"%@\n%@", NSStringFromSelector(_cmd), [NSThread callStackSymbols]);
     [[iTermWindowOcclusionChangeMonitor sharedInstance] invalidateCachedOcclusion];
     [super orderBack:sender];
 }
@@ -313,6 +342,18 @@ ITERM_WEAKLY_REFERENCEABLE
 
     NSArray *orderedWindows = [[NSApplication sharedApplication] orderedWindows];
     NSUInteger myIndex = [orderedWindows indexOfObject:self];
+    NSRect myFrame = [self frame];
+    double onScreenFraction = 0;
+    const double myArea = myFrame.size.width * myFrame.size.height;
+    if (myArea >= 1) {
+        for (NSScreen *screen in [NSScreen screens]) {
+            const NSRect screenFrame = screen.frame;
+            const NSRect onscreenFrame = NSIntersectionRect(myFrame, screenFrame);
+            const CGFloat onscreenArea = (onscreenFrame.size.width * onscreenFrame.size.height);
+            onScreenFraction += onscreenArea / myArea;
+        }
+    }
+    const double offscreenFraction = MAX(MIN(1, 1 - onScreenFraction), 0);
     CGFloat totalOcclusion = 0;
     if (myIndex != 0 && myIndex != NSNotFound) {
         const int kRows = 3;
@@ -322,7 +363,7 @@ ITERM_WEAKLY_REFERENCEABLE
             double occlusion;
         } OcclusionPart;
         OcclusionPart parts[kRows][kCols];
-        NSRect myFrame = [self frame];
+
         NSSize partSize = NSMakeSize(myFrame.size.width / kCols, myFrame.size.height / kRows);
         for (int y = 0; y < kRows; y++) {
             for (int x = 0; x < kCols; x++) {
@@ -337,7 +378,7 @@ ITERM_WEAKLY_REFERENCEABLE
 
         // This loop iterates over each window in front of this one and measures
         // how much of it intersects each part of this one (a part is one 9th of
-        // the window, as divded into a 3x3 grid). For each part, an occlusion
+        // the window, as divided into a 3x3 grid). For each part, an occlusion
         // fraction is tracked, which is the fraction of that part which is covered
         // by another window. It's approximate because it's the maximum occlusion
         // for that part by all other windows, so it could be too low (if two
@@ -367,13 +408,16 @@ ITERM_WEAKLY_REFERENCEABLE
                     totalOcclusion += parts[y][x].occlusion / (kRows * kCols);
                 }
             }
+            totalOcclusion = MAX(totalOcclusion, offscreenFraction);
             if (totalOcclusion > 0.99) {
                 totalOcclusion = 1;
                 break;
             }
         }
+    } else {
+        totalOcclusion = offscreenFraction;
     }
-    
+
     _totalOcclusionCacheTime = [NSDate timeIntervalSinceReferenceDate];
     _cachedTotalOcclusion = totalOcclusion;
     return totalOcclusion;
@@ -397,6 +441,77 @@ ITERM_WEAKLY_REFERENCEABLE
     return NSWindowTabbingModeDisallowed;
 }
 
+- (void)setFrame:(NSRect)frameRect display:(BOOL)flag {
+    DLog(@"setFrame:%@ display:%@ maxy=%@ from\n%@",
+         NSStringFromRect(frameRect), @(flag), @(NSMaxY(frameRect)),
+         [NSThread callStackSymbols]);
+    [super setFrame:frameRect display:flag];
+}
+
+- (void)setFrameOrigin:(NSPoint)point {
+    DLog(@"Set frame origin to %@", NSStringFromPoint(point));
+    [super setFrameOrigin:point];
+    DLog(@"Frame maxy=%@ now", @(NSMaxY(self.frame)));
+}
+
+#if ENABLE_COMPACT_WINDOW_HACK
+- (BOOL)isCompact {
+    return YES;
+}
+
++ (Class)frameViewClassForStyleMask:(NSUInteger)windowStyle {
+    return [iTermThemeFrame class] ?: [super frameViewClassForStyleMask:windowStyle];
+}
+
+// https://chromium.googlesource.com/chromium/src/+/refs/tags/73.0.3683.86/ui/views_bridge_mac/native_widget_mac_nswindow.mm#169
+// The base implementation returns YES if the window's frame view is a custom
+// class, which causes undesirable changes in behavior. AppKit NSWindow
+// subclasses are known to override it and return NO.
+//
+// In particular, this fixes issue 8478 (traffic light buttons vertically off-center in native full screen).
+- (BOOL)_usesCustomDrawing {
+    return NO;
+}
+#else
+- (BOOL)isCompact {
+    return NO;
+}
+#endif
+
+- (NSColor *)it_terminalWindowDecorationBackgroundColor {
+    return [self.ptyDelegate terminalWindowDecorationBackgroundColor];
+}
+
+- (id<PSMTabStyle>)it_tabStyle {
+    return [self.ptyDelegate terminalWindowTabStyle];
+}
+
+- (NSColor *)it_terminalWindowDecorationTextColorForBackgroundColor:(NSColor *)backgroundColor {
+    return [self.ptyDelegate terminalWindowDecorationTextColorForBackgroundColor:(NSColor *)backgroundColor];
+}
+
+- (NSColor *)it_terminalWindowDecorationControlColor {
+    return [self.ptyDelegate terminalWindowDecorationControlColor];
+}
+
+- (BOOL)it_terminalWindowUseMinimalStyle {
+    return [self.ptyDelegate terminalWindowUseMinimalStyle];
+}
+
+- (void)beginSheet:(NSWindow *)sheetWindow completionHandler:(void (^ _Nullable)(NSModalResponse))handler {
+    self.it_openingSheet = self.it_openingSheet + 1;
+    [super beginSheet:sheetWindow completionHandler:handler];
+    self.it_openingSheet = self.it_openingSheet - 1;
+}
+
+- (void)it_setNeedsInvalidateShadow {
+    _needsInvalidateShadow = YES;
+    dispatch_async(dispatch_get_main_queue(), ^{
+        _needsInvalidateShadow = NO;
+        [self invalidateShadow];
+    });
+}
+
 NS_ASSUME_NONNULL_END
 
 @end
